#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################
from typing import Optional

import torch  # noqa: F401
from torch import nn, Tensor

from .functional import neighborhood_attention_generic
from .types import (
    CausalArg1DTypeOrDed,
    CausalArg2DTypeOrDed,
    CausalArg3DTypeOrDed,
    CausalArgTypeOrDed,
    Dimension1DTypeOrDed,
    Dimension2DTypeOrDed,
    Dimension3DTypeOrDed,
    DimensionTypeOrDed,
)
from .utils.checks import check_all_args


class NeighborhoodAttentionGeneric(nn.Module):
    def __init__(
        self,
        na_dim: int,
        embed_dim: int,
        num_heads: int,
        kernel_size: DimensionTypeOrDed,
        stride: DimensionTypeOrDed = 1,
        dilation: DimensionTypeOrDed = 1,
        is_causal: CausalArgTypeOrDed = False,
        qkv_bias: bool = True,
        qk_scale: Optional[float] = None,
        proj_drop: float = 0.0,
    ):
        super().__init__()
        kernel_size, stride, dilation, is_causal = check_all_args(
            na_dim, kernel_size, stride, dilation, is_causal
        )

        if embed_dim % num_heads != 0:
            raise ValueError(
                "Number of attention heads must evenly divide embedding dimension, "
                f"got {embed_dim=}, {num_heads=}."
            )

        self.na_dim = na_dim
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.head_dim = self.embed_dim // self.num_heads
        self.scale = qk_scale or self.head_dim**-0.5
        self.kernel_size = kernel_size
        self.stride = stride
        self.dilation = dilation
        self.is_causal = is_causal

        self.expected_input_tensor_rank = self.na_dim + 2  # batch, embedding dim

        self.qkv = nn.Linear(self.embed_dim, self.embed_dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: Tensor) -> Tensor:
        if x.dim() != self.expected_input_tensor_rank:
            raise ValueError(
                f"NeighborhoodAttention{self.na_dim}D expected a tensor with rank "
                f"{self.expected_input_tensor_rank} ({self.na_dim} for token layout, 1 for batch, "
                f"1 for embedding dimension), got {x.dim()=}."
            )

        B, *input_shape, C = x.shape

        if C != self.embed_dim:
            raise ValueError(
                f"Expected embedding dimension {self.embed_dim}, got {C} ({x.shape=})."
            )

        # 3, batch, *input_shape, heads, head_dim
        permutation = (
            [self.na_dim + 1, 0]
            + [x + 1 for x in range(self.na_dim)]
            + [self.na_dim + 2, self.na_dim + 3]
        )
        qkv = (
            self.qkv(x)
            .reshape(B, *input_shape, 3, self.num_heads, self.head_dim)
            .permute(*permutation)
        )
        q, k, v = qkv[0], qkv[1], qkv[2]
        x = neighborhood_attention_generic(
            q,
            k,
            v,
            kernel_size=self.kernel_size,
            stride=self.stride,
            dilation=self.dilation,
            is_causal=self.is_causal,
            scale=self.scale,
        )
        x = x.reshape(B, *input_shape, C)

        return self.proj_drop(self.proj(x))

    def extra_repr(self) -> str:
        return (
            f"head_dim={self.head_dim}, num_heads={self.num_heads}, "
            + f"kernel_size={self.kernel_size}, "
            + f"stride={self.stride}, "
            + f"dilation={self.dilation}, "
            + f"is_causal={self.is_causal}"
        )


class NeighborhoodAttention1D(NeighborhoodAttentionGeneric):
    """
    1-D Neighborhood Attention torch module.

    Performs QKV and output linear projections in addition to the [na1d][natten.na1d] operation.

    Args:
        embed_dim: Embedding dimension size (a.k.a. number of channels, latent size).
            !!! note
                This is not `head_dim`. It's `head_dim * num_heads`.

        num_heads: Number of attention heads.

        kernel_size (Tuple[int] | int): Neighborhood window (kernel) size.

            !!! note
                `kernel_size` must be smaller than or equal to `seqlen`.

        stride (Tuple[int] | int): Sliding window step size. Defaults to `1` (standard sliding
            window).

            !!! note
                `stride` must be smaller than or equal to `kernel_size`.
                When `stride == kernel_size`, there will be no overlap between sliding windows,
                which is equivalent to blocked attention (a.k.a.
                [window self attention](https://arxiv.org/abs/2103.14030)).

        dilation (Tuple[int] | int): Dilation step size. Defaults to `1` (standard sliding window).

            !!! note
                The product of `dilation` and `kernel_size` must be smaller than or equal to
                `seqlen`.

        is_causal (Tuple[bool] | bool): Toggle causal masking. Defaults to `False`
            (bi-directional).

        qkv_bias: Enable bias in the QKV linear projection.

        qk_scale: Attention scale. Defaults to `head_dim ** -0.5`.

        proj_drop: Dropout score for projection layer. Defaults is `0.0` (no dropout).

    Example:
        ```python3
        import torch
        from natten import NeighborhoodAttention1D

        num_heads = 4
        head_dim = 128
        embed_dim = num_heads * head_dim

        model = NeighborhoodAttention1D(
            embed_dim=embed_dim,
            num_heads=num_heads,
            kernel_size=2048,
            stride=2,
            dilation=4,
            is_causal=True
        )

        batch = 1
        seqlen = 4096 # (1)!

        x = torch.randn(batch, seqlen, embed_dim) # (2)!
        y = model(x) # (3)!
        ```

        1. Tokens are arranged in a sequential layout of size 4096, to which we apply a
            kernel size of 2048, stride 2, dilation 4, and apply causal masking.

        2. `x.shape == [1, 4096, 512]`
        3. `y.shape == [1, 4096, 512]`
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        kernel_size: Dimension1DTypeOrDed,
        stride: Dimension1DTypeOrDed = 1,
        dilation: Dimension1DTypeOrDed = 1,
        is_causal: CausalArg1DTypeOrDed = False,
        qkv_bias: bool = True,
        qk_scale: Optional[float] = None,
        proj_drop: float = 0.0,
    ):
        super().__init__(
            na_dim=1,
            embed_dim=embed_dim,
            num_heads=num_heads,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            is_causal=is_causal,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            proj_drop=proj_drop,
        )


class NeighborhoodAttention2D(NeighborhoodAttentionGeneric):
    """
    2-D Neighborhood Attention torch module.

    Performs QKV and output linear projections in addition to the [na2d][natten.na2d] operation.

    Args:
        embed_dim: Embedding dimension size (a.k.a. number of channels, latent size).
            !!! note
                This is not `head_dim`. It's `head_dim * num_heads`.

        num_heads: Number of attention heads.

        kernel_size (Tuple[int, int] | int): Neighborhood window (kernel) size/shape. If an
            integer, it will be repeated for all 2 dimensions. For example `kernel_size=3` is
            reinterpreted as `kernel_size=(3, 3)`.

            !!! note
                `kernel_size` must be smaller than or equal to token layout shape (`(X, Y)`) along
                every dimension.

        stride (Tuple[int, int] | int): Sliding window step size/shape. Defaults to `1` (standard
            sliding window). If an integer, it will be repeated for all 2 dimensions. For example
            `stride=2` is reinterpreted as `stride=(2, 2)`.

            !!! note
                `stride` must be smaller than or equal to `kernel_size` along every dimension.
                When `stride == kernel_size`, there will be no overlap between sliding windows,
                which is equivalent to blocked attention (a.k.a.
                [window self attention](https://arxiv.org/abs/2103.14030)).

        dilation (Tuple[int, int] | int): Dilation step size/shape. Defaults to `1` (standard
            sliding window). If an integer, it will be repeated for all 2 dimensions. For example
            `dilation=4` is reinterpreted as `dilation=(4, 4)`.

            !!! note
                The product of `dilation` and `kernel_size` must be smaller than or equal to
                token layout shape (`(X, Y)`) along every dimension.

        is_causal (Tuple[bool, bool] | bool): Toggle causal masking. Defaults to `False`
            (bi-directional). If a boolean, it will be repeated for all 2 dimensions. For example
            `is_causal=True` is reinterpreted as `is_causal=(True, True)`.

        qkv_bias: Enable bias in the QKV linear projection.

        qk_scale: Attention scale. Defaults to `head_dim ** -0.5`.

        proj_drop: Dropout score for projection layer. Defaults is `0.0` (no dropout).

    Example:
        ```python3
        import torch
        from natten import NeighborhoodAttention2D

        num_heads = 4
        head_dim = 128
        embed_dim = num_heads * head_dim

        model = NeighborhoodAttention2D(
            embed_dim=embed_dim,
            num_heads=num_heads,
            kernel_size=(8, 16),
            stride=(1, 2),
            dilation=(2, 1),
            is_causal=False
        )

        batch = 1
        token_layout_shape = (16, 32) # (1)!

        x = torch.randn(batch, *token_layout_shape, embed_dim) # (2)!
        y = model(x) # (3)!
        ```

        1. Tokens are arranged in a 16 x 32 layout, to which we apply a
            kernel size of 8 x 16,
            stride 1 x 2,
            and dilation 2 x 1.

        2. `x.shape == [1, 16, 32, 512]`
        3. `y.shape == [1, 16, 32, 512]`
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        kernel_size: Dimension2DTypeOrDed,
        stride: Dimension2DTypeOrDed = 1,
        dilation: Dimension2DTypeOrDed = 1,
        is_causal: CausalArg2DTypeOrDed = False,
        qkv_bias: bool = True,
        qk_scale: Optional[float] = None,
        proj_drop: float = 0.0,
    ):
        super().__init__(
            na_dim=2,
            embed_dim=embed_dim,
            num_heads=num_heads,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            is_causal=is_causal,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            proj_drop=proj_drop,
        )


class NeighborhoodAttention3D(NeighborhoodAttentionGeneric):
    """
    3-D Neighborhood Attention torch module.

    Performs QKV and output linear projections in addition to the [na3d][natten.na3d] operation.

    Args:
        embed_dim: Embedding dimension size (a.k.a. number of channels, latent size).
            !!! note
                This is not `head_dim`. It's `head_dim * num_heads`.

        num_heads: Number of attention heads.

        kernel_size (Tuple[int, int, int] | int): Neighborhood window (kernel) size/shape. If an
            integer, it will be repeated for all 3 dimensions. For example `kernel_size=3` is
            reinterpreted as `kernel_size=(3, 3, 3)`.

            !!! note
                `kernel_size` must be smaller than or equal to token layout shape (`(X, Y, Z)`)
                along every dimension.

        stride (Tuple[int, int, int] | int): Sliding window step size/shape. Defaults to `1`
            (standard sliding window). If an integer, it will be repeated for all 3 dimensions.
            For example `stride=2` is reinterpreted as `stride=(2, 2, 2)`.

            !!! note
                `stride` must be smaller than or equal to `kernel_size` along every dimension.
                When `stride == kernel_size`, there will be no overlap between sliding windows,
                which is equivalent to blocked attention (a.k.a.
                [window self attention](https://arxiv.org/abs/2103.14030)).

        dilation (Tuple[int, int, int] | int): Dilation step size/shape. Defaults to `1` (standard
            sliding window). If an integer, it will be repeated for all 3 dimensions. For example
            `dilation=4` is reinterpreted as `dilation=(4, 4, 4)`.

            !!! note
                The product of `dilation` and `kernel_size` must be smaller than or equal to
                token layout shape (`(X, Y, Z)`) along every dimension.

        is_causal (Tuple[bool, bool, bool] | bool): Toggle causal masking. Defaults to `False`
            (bi-directional). If a boolean, it will be repeated for all 3 dimensions. For example
            `is_causal=True` is reinterpreted as `is_causal=(True, True, True)`.

        qkv_bias: Enable bias in the QKV linear projection.

        qk_scale: Attention scale. Defaults to `head_dim ** -0.5`.

        proj_drop: Dropout score for projection layer. Defaults is `0.0` (no dropout).

    Example:
        ```python3
        import torch
        from natten import NeighborhoodAttention3D

        num_heads = 4
        head_dim = 128
        embed_dim = num_heads * head_dim

        model = NeighborhoodAttention3D(
            embed_dim=embed_dim,
            num_heads=num_heads,
            kernel_size=(4, 8, 12),
            stride=(1, 1, 4),
            dilation=(1, 2, 1),
            is_causal=(True, False, False)
        )

        batch = 1
        token_layout_shape = (12, 16, 20) # (1)!

        x = torch.randn(batch, *token_layout_shape, embed_dim) # (2)!
        y = model(x) # (3)!
        ```

        1. Tokens are arranged in a 12 x 16 x 20 layout, to which we apply a
            kernel size of 4 x 8 x 12,
            stride 1 x 1 x 4,
            dilation 1 x 2 x 1, and apply causal masking to the left-most dimension (12).

        2. `x.shape == [1, 12, 16, 20, 512]`
        3. `y.shape == [1, 12, 16, 20, 512]`
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        kernel_size: Dimension3DTypeOrDed,
        stride: Dimension3DTypeOrDed = 1,
        dilation: Dimension3DTypeOrDed = 1,
        is_causal: CausalArg3DTypeOrDed = False,
        qkv_bias: bool = True,
        qk_scale: Optional[float] = None,
        proj_drop: float = 0.0,
    ):
        super().__init__(
            na_dim=3,
            embed_dim=embed_dim,
            num_heads=num_heads,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            is_causal=is_causal,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            proj_drop=proj_drop,
        )



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################
import functools
from typing import Optional, Tuple, Union

import torch
from torch import Tensor

from torch.amp import custom_bwd, custom_fwd
from torch.autograd import Function

amp_fwd = functools.partial(custom_fwd, device_type="cuda")
amp_bwd = functools.partial(custom_bwd, device_type="cuda")

from .._libnatten import (
    na1d_backward,
    na1d_forward,
    na2d_backward,
    na2d_forward,
    na3d_backward,
    na3d_forward,
)
from ..types import (
    CausalArg1DTypeOrDed,
    CausalArg2DTypeOrDed,
    CausalArg3DTypeOrDed,
    CausalArgType,
    CausalArgTypeOrDed,
    CutlassFnaBackwardConfigType,
    CutlassFnaForwardConfigType,
    Dimension1DType,
    Dimension1DTypeOrDed,
    Dimension2DType,
    Dimension2DTypeOrDed,
    Dimension3DType,
    Dimension3DTypeOrDed,
    DimensionType,
    DimensionTypeOrDed,
    NoneType,
)
from ..utils import log
from ..utils.checks import check_all_args, check_args_against_input, na_tensor_checks

from .configs.checks import can_run_cutlass_fna
from .configs.cutlass import (
    check_cutlass_fna_backward_config,
    check_cutlass_fna_forward_config,
)

logger = log.get_logger(__name__)


def make_cutlass_fna_autograd_fn(na_dim):
    assert na_dim in [1, 2, 3]

    FORWARD_OPS = {
        1: na1d_forward,
        2: na2d_forward,
        3: na3d_forward,
    }

    BACKWARD_OPS = {
        1: na1d_backward,
        2: na2d_backward,
        3: na3d_backward,
    }

    class CutlassFnaGenericAutogradFn(Function):
        @staticmethod
        @amp_fwd
        def forward(
            ctx,
            query: Tensor,
            key: Tensor,
            value: Tensor,
            kernel_size: DimensionType,
            stride: DimensionType,
            dilation: DimensionType,
            is_causal: CausalArgType,
            scale: float,
            forward_config: CutlassFnaForwardConfigType,
            backward_config: CutlassFnaBackwardConfigType,
        ) -> Tuple[Tensor, Tensor]:
            kernel_size, stride, dilation, is_causal = check_all_args(
                na_dim, kernel_size, stride, dilation, is_causal
            )

            query = query.contiguous()
            key = key.contiguous()
            value = value.contiguous()

            assert query.dim() == value.dim() == 3 + na_dim
            assert query.shape[0] == value.shape[0]
            assert query.shape[-2] == value.shape[-2]

            output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
            output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

            # TODO: logsumexp should be conditional
            logsumexp = torch.empty(
                query.shape[:-1], dtype=torch.float32, device=query.device
            )

            q_tile_shape, kv_tile_shape = forward_config
            FORWARD_OPS[na_dim](
                output,
                query,
                key,
                value,
                logsumexp,
                kernel_size,
                stride,
                dilation,
                is_causal,
                scale,
                q_tile_shape,
                kv_tile_shape,
            )

            ctx.save_for_backward(query, key, value, logsumexp, output)
            ctx.kernel_size = kernel_size
            ctx.stride = stride
            ctx.dilation = dilation
            ctx.is_causal = is_causal
            ctx.scale = scale
            ctx.backward_config = backward_config

            return output, logsumexp

        @staticmethod
        @amp_bwd
        def backward(ctx, grad_out: Tensor, grad_lse: Tensor) -> Tuple[
            Tensor,
            Tensor,
            Tensor,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
        ]:
            query, key, value, logsumexp, output = ctx.saved_tensors
            d_output = grad_out.contiguous()
            d_query = torch.empty_like(query)
            d_key = torch.empty_like(key)
            d_value = torch.empty_like(value)

            q_tile_shape, k_tile_shape, kv_splits, compute_delta_with_pt = (
                ctx.backward_config
            )

            num_kv_splits = kv_splits
            if (
                any([kv_split > 1 for kv_split in kv_splits])
                and torch.are_deterministic_algorithms_enabled()
            ):
                num_kv_splits = tuple(1 for _ in range(len(kv_splits)))
                logger.warning(
                    "You enabled PyTorch's deterministic mode, but tried to train with FNA's KV "
                    "parallelism, which is non-deterministic. "
                    f"Overriding {kv_splits} to {num_kv_splits}."
                )

            BACKWARD_OPS[na_dim](
                d_query,
                d_key,
                d_value,
                query,
                key,
                value,
                output,
                d_output,
                logsumexp,
                ctx.kernel_size,
                ctx.stride,
                ctx.dilation,
                ctx.is_causal,
                ctx.scale,
                q_tile_shape,
                k_tile_shape,
                num_kv_splits,
                compute_delta_with_pt,
            )

            return d_query, d_key, d_value, None, None, None, None, None, None, None

    return CutlassFnaGenericAutogradFn


CutlassFna1DAutogradFn = make_cutlass_fna_autograd_fn(1)
CutlassFna2DAutogradFn = make_cutlass_fna_autograd_fn(2)
CutlassFna3DAutogradFn = make_cutlass_fna_autograd_fn(3)


CutlassFNAAutogradFns = {
    1: CutlassFna1DAutogradFn,
    2: CutlassFna2DAutogradFn,
    3: CutlassFna3DAutogradFn,
}


def cutlass_fna_generic(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: DimensionTypeOrDed,
    stride: DimensionTypeOrDed = 1,
    dilation: DimensionTypeOrDed = 1,
    is_causal: Optional[CausalArgTypeOrDed] = False,
    scale: Optional[float] = None,
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
    backward_q_tile_shape: Optional[DimensionType] = None,
    backward_kv_tile_shape: Optional[DimensionType] = None,
    backward_kv_splits: Optional[DimensionType] = None,
    backward_use_pt_reduction: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:

    na_tensor_checks(query, key, value, must_match_head_dims=False)

    assert can_run_cutlass_fna(query, key, value, raise_error=True)

    na_dim = query.dim() - 3  # batch, heads, head_dim

    assert na_dim in [1, 2, 3]

    kernel_size, stride, dilation, is_causal = check_all_args(
        na_dim, kernel_size, stride, dilation, is_causal
    )

    check_args_against_input(
        query,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
    )

    forward_config = check_cutlass_fna_forward_config(
        input_tensor=query if value.shape[-1] <= query.shape[-1] else value,
        dilation=dilation,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
    )

    backward_config = check_cutlass_fna_backward_config(
        input_tensor=key if value.shape[-1] <= key.shape[-1] else value,
        dilation=dilation,
        q_tile_shape=backward_q_tile_shape,
        kv_tile_shape=backward_kv_tile_shape,
        kv_splits=backward_kv_splits,
        use_pt_reduction=backward_use_pt_reduction,
    )

    scale = scale or query.shape[-1] ** -0.5

    output, lse = CutlassFNAAutogradFns[na_dim].apply(
        query,
        key,
        value,
        kernel_size,
        stride,
        dilation,
        is_causal,
        scale,
        forward_config,
        backward_config,
    )

    if return_lse:
        return output, lse

    return output


def na1d_cutlass_fna(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension1DTypeOrDed,
    stride: Dimension1DTypeOrDed = 1,
    dilation: Dimension1DTypeOrDed = 1,
    is_causal: Optional[CausalArg1DTypeOrDed] = False,
    scale: Optional[float] = None,
    q_tile_shape: Optional[Dimension1DType] = None,
    kv_tile_shape: Optional[Dimension1DType] = None,
    backward_q_tile_shape: Optional[Dimension1DType] = None,
    backward_kv_tile_shape: Optional[Dimension1DType] = None,
    backward_kv_splits: Optional[Dimension1DType] = None,
    backward_use_pt_reduction: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    return cutlass_fna_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        backward_q_tile_shape=backward_q_tile_shape,
        backward_kv_tile_shape=backward_kv_tile_shape,
        backward_kv_splits=backward_kv_splits,
        backward_use_pt_reduction=backward_use_pt_reduction,
        return_lse=return_lse,
    )


def na2d_cutlass_fna(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension2DTypeOrDed,
    stride: Dimension2DTypeOrDed = 1,
    dilation: Dimension2DTypeOrDed = 1,
    is_causal: Optional[CausalArg2DTypeOrDed] = False,
    scale: Optional[float] = None,
    q_tile_shape: Optional[Dimension2DType] = None,
    kv_tile_shape: Optional[Dimension2DType] = None,
    backward_q_tile_shape: Optional[Dimension2DType] = None,
    backward_kv_tile_shape: Optional[Dimension2DType] = None,
    backward_kv_splits: Optional[Dimension2DType] = None,
    backward_use_pt_reduction: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    return cutlass_fna_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        backward_q_tile_shape=backward_q_tile_shape,
        backward_kv_tile_shape=backward_kv_tile_shape,
        backward_kv_splits=backward_kv_splits,
        backward_use_pt_reduction=backward_use_pt_reduction,
        return_lse=return_lse,
    )


def na3d_cutlass_fna(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension3DTypeOrDed,
    stride: Dimension3DTypeOrDed = 1,
    dilation: Dimension3DTypeOrDed = 1,
    is_causal: Optional[CausalArg3DTypeOrDed] = False,
    scale: Optional[float] = None,
    q_tile_shape: Optional[Dimension3DType] = None,
    kv_tile_shape: Optional[Dimension3DType] = None,
    backward_q_tile_shape: Optional[Dimension3DType] = None,
    backward_kv_tile_shape: Optional[Dimension3DType] = None,
    backward_kv_splits: Optional[Dimension3DType] = None,
    backward_use_pt_reduction: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    return cutlass_fna_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        backward_q_tile_shape=backward_q_tile_shape,
        backward_kv_tile_shape=backward_kv_tile_shape,
        backward_kv_splits=backward_kv_splits,
        backward_use_pt_reduction=backward_use_pt_reduction,
        return_lse=return_lse,
    )



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import argparse
import math
import time
from collections.abc import Sequence
from functools import partial
from typing import List, Optional, Union

import torch

from . import (
    allow_flex_compile,
    set_memory_usage_preference,
    use_kv_parallelism_in_fused_na,
)

from .profiling_utils import (
    generate_problem,
    print_table,
    profile_fmha_with_torch,
    profile_na_with_torch,
)
from .profiling_utils.dry_run import dry_run as dry_run_fn, optimize as optimize_fn
from .profiling_utils.formatting import Result
from .profiling_utils.problem import Problem
from .types import CausalArgType, DimensionType, KernelSchedule
from .utils import log

logger = log.get_logger(__name__)

__all__ = []  # type: ignore

SUPPORTED_DTYPES = ["fp32", "bf16", "fp16"]
NATTEN_BACKENDS = ["cutlass-fna", "blackwell-fna", "hopper-fna", "flex-fna"]
NATTEN_FMHA_BACKENDS = ["cutlass-fmha", "blackwell-fmha", "hopper-fmha", "flex-fmha"]
SDPA_BACKENDS = ["xformers", "cudnn", "fav2"]

SCHEDULE_MAP = {
    "non": KernelSchedule.NonPersistent,
    "coop": KernelSchedule.WarpSpecializedCooperative,
    "pp": KernelSchedule.WarpSpecializedPingpong,
}

DEFAULT_DTYPE = "fp16"
DEFAULT_BACKEND = "fna"


def do_profile(
    problem: Problem,
    backend: Optional[str] = None,
    fmha_backend: Optional[str] = None,
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
    backward_q_tile_shape: Optional[DimensionType] = None,
    backward_kv_tile_shape: Optional[DimensionType] = None,
    persistent: bool = True,
    torch_compile: bool = False,
    kernel_schedule: Optional[Union[str, KernelSchedule]] = None,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
    backward_q_tile_size: Optional[int] = None,
    backward_kv_tile_size: Optional[int] = None,
    warmup_steps: int = 5,
    backprop: bool = False,
    max_retries: int = 5,
    fail_on_capture_error: bool = True,
):
    if backend is None or backend in NATTEN_BACKENDS:
        func = partial(
            profile_na_with_torch,
            backend=backend,
            fmha_backend=fmha_backend,
            q_tile_shape=q_tile_shape,
            kv_tile_shape=kv_tile_shape,
            backward_q_tile_shape=backward_q_tile_shape,
            backward_kv_tile_shape=backward_kv_tile_shape,
            run_persistent_kernel=persistent,
            torch_compile=torch_compile,
            kernel_schedule=kernel_schedule,
            q_tile_size=q_tile_size,
            kv_tile_size=kv_tile_size,
            backward_q_tile_size=backward_q_tile_size,
            backward_kv_tile_size=backward_kv_tile_size,
        )

    elif backend in SDPA_BACKENDS:
        func = partial(profile_fmha_with_torch, backend=backend)

    else:
        raise NotImplementedError(f"Unrecognized backend {backend}.")

    logged_ops: Optional[List[Result]] = func(
        problem=problem,
        warmup_steps=warmup_steps,
        disable_backward=not backprop,
    )

    # Temp solution for an intermittent bug where trace is empty (kernel launch, device synch, and
    # all that stuff is still there, but for some reason no kernels.... I've seen this happen in
    # many profiling tools, so I think it may be a driver bug or something.)
    retries = 0
    while logged_ops is None or len(logged_ops) < 1:
        logger.debug(
            "Profiler trace was empty, retrying after device sync and short sleep."
        )
        torch.cuda.synchronize()
        time.sleep(1)
        torch.cuda.synchronize()

        retries += 1
        logger.debug(f"Retrying [attempt {retries} / {max_retries}] ...")

        logged_ops = func(
            problem=problem,
            warmup_steps=warmup_steps,
            disable_backward=not backprop,
            debug_report_prof_result=True,
        )

        if retries >= max_retries:
            break

    if fail_on_capture_error and (logged_ops is None or len(logged_ops) < 1):
        raise RuntimeError(
            f"Nothing was measured during profiling after {max_retries} retries. "
            "Please open an issue and copy paste the following information:\n"
            f"{problem=},\n"
            f"{backend=},\n"
            f"{fmha_backend=},\n"
            f"{q_tile_shape=},\n"
            f"{kv_tile_shape=},\n"
            f"{backward_q_tile_shape=},\n"
            f"{backward_kv_tile_shape=},\n"
            f"{persistent=},\n"
            f"{torch_compile=},\n"
            f"{kernel_schedule=},\n"
            f"{warmup_steps=},\n"
            f"{backprop=},\n"
            f"{logged_ops=},\n"
        )

    return logged_ops


def check_input_size(input_size: Sequence) -> DimensionType:
    if input_size is None:
        raise ValueError(f"Input size is a required argument, got {input_size}.")

    if len(input_size) < 1 or len(input_size) > 3:
        raise ValueError(
            f"Input size must be a tuple of either 1, 2 or 3 integers, got {len(input_size)}."
        )

    return tuple(x for x in input_size)


def check_window_size(
    input_size: Sequence, window_size: Optional[Sequence] = None
) -> DimensionType:
    assert isinstance(input_size, Sequence) and len(input_size) in [1, 2, 3]
    na_dim = len(input_size)

    if window_size is None:
        return tuple(x for x in input_size)  # type: ignore[return-value]

    if len(window_size) != na_dim:
        raise ValueError(
            f"Invalid window size {window_size} for input size {input_size}. "
            "They must match in number of elements."
        )

    for w, x in zip(window_size, input_size):
        if w > x:
            raise ValueError(
                f"Invalid window size {window_size} for input size {input_size}. "
                "Window size cannot be greater than input size along any dimension."
            )

    return tuple(w for w in window_size)


def check_dilation(
    input_size: Sequence, window_size: Sequence, dilation: Optional[Sequence] = None
) -> DimensionType:
    assert isinstance(input_size, Sequence) and len(input_size) in [1, 2, 3]
    assert isinstance(window_size, Sequence) and len(window_size) == len(input_size)
    na_dim = len(input_size)

    if dilation is None:
        return tuple(1 for _ in range(na_dim))  # type: ignore[return-value]

    if len(dilation) != na_dim:
        raise ValueError(
            f"Invalid dilation {dilation} for input size {input_size}. "
            "They must match in number of elements."
        )

    for d, w, x in zip(dilation, window_size, input_size):
        if d * w > x:
            raise ValueError(
                f"Invalid dilation {dilation} for input size {input_size} and window size {window_size}. "
                "The product of window size and dilation cannot be greater than input size along any dimension."
            )

    return tuple(d for d in dilation)


def check_stride(
    input_size: Sequence,
    window_size: Sequence,
    dilation: Sequence,
    stride: Optional[Sequence] = None,
) -> DimensionType:
    assert isinstance(input_size, Sequence) and len(input_size) in [1, 2, 3]
    assert isinstance(window_size, Sequence) and len(window_size) == len(input_size)
    assert isinstance(dilation, Sequence) and len(dilation) == len(input_size)
    na_dim = len(input_size)

    if stride is None:
        return tuple(1 for _ in range(na_dim))  # type: ignore[return-value]

    if len(stride) != na_dim:
        raise ValueError(
            f"Invalid stride {stride} for input size {input_size}. "
            "They must match in number of elements."
        )

    for s, d, w, x in zip(stride, dilation, window_size, input_size):
        if s > w:
            raise ValueError(
                f"Invalid stride {stride} for window size {window_size}. "
                "Stride cannot be larger than window size along any dimension."
            )

    return tuple(s for s in stride)


def check_causal(
    input_size: Sequence, causal: Optional[Sequence] = None
) -> CausalArgType:
    assert isinstance(input_size, Sequence) and len(input_size) in [1, 2, 3]
    na_dim = len(input_size)

    if causal is None:
        return tuple(False for _ in range(na_dim))  # type: ignore[return-value]

    return tuple(c for c in causal)


def get_args():
    parser = argparse.ArgumentParser(
        description="NATTEN profiling toolkit. Profile different scenarios with NATTEN and "
        "see accurate measurements of operation runtime under different backends and configurations."
    )

    parser.add_argument(
        "-b", "--batch-size", type=int, default=1, help="QKV batch size."
    )
    parser.add_argument(
        "-n", "--heads", type=int, default=1, help="QKV number of heads (no GQA/MQA)."
    )
    parser.add_argument(
        "-d", "--dim", type=int, default=64, help="QK (and optionally V) head dim."
    )
    parser.add_argument(
        "--dim-value",
        type=int,
        default=None,
        help="Head dim for the V tensor, if different from Q and K. Defaults to the value of "
        "`-d`/`--dim`.",
    )

    parser.add_argument(
        "-i",
        "--input-size",
        type=int,
        nargs="+",
        help="QKV token layout shape (i.e. sequence length in 1-D, "
        "height and width in 2-D, depth, height, and width in 3-D.)",
    )

    parser.add_argument(
        "-w",
        "--window-size",
        type=int,
        nargs="*",
        help="Neighborhood attention window size (shape), also referred to "
        "as kernel size. This must be a tuple with the same number of elements "
        "as in --input-size. Defaults to --input-size (self attention).",
    )

    parser.add_argument(
        "-s",
        "--stride",
        type=int,
        nargs="*",
        help="Neighborhood attention stride values. This must be a tuple with the "
        "same number of elements as in --input-size. Defaults to 1s (standard "
        "sliding window).",
    )

    parser.add_argument(
        "--dilation",
        type=int,
        nargs="*",
        help="Neighborhood attention dilation values. This must be a tuple with the "
        "same number of elements as in --input-size. Defaults to 1s (standard "
        "sliding window).",
    )

    parser.add_argument(
        "-c",
        "--causal",
        type=bool,
        nargs="*",
        help="Causal masking values. This must be a boolean tuple with the "
        "same number of elements as in --input-size. Defaults to all `False`s ("
        "bi-directional in all dimensions).",
    )

    parser.add_argument(
        "--dtype",
        type=str,
        default=DEFAULT_DTYPE,
        choices=SUPPORTED_DTYPES,
        help=f"Element (data) type. Choices: {', '.join(SUPPORTED_DTYPES)}",
    )

    parser.add_argument(
        "--backprop",
        action="store_true",
        help="Profile backward pass as well as forward pass.",
    )

    parser.add_argument(
        "--add-kv",
        type=int,
        default=0,
        help="Number of additional KV tokens. Defaults to 0.",
    )

    parser.add_argument(
        "--backend",
        type=str,
        default=None,
        choices=NATTEN_BACKENDS + SDPA_BACKENDS,
        help="Backend / kernel to run."
        "Choices: "
        f"NATTEN backends: {', '.join(NATTEN_BACKENDS)}.\n"
        f"Torch SDPA backends (can only perform self attention): {', '.join(SDPA_BACKENDS)}.",
    )

    parser.add_argument(
        "--fmha-backend",
        type=str,
        default=None,
        choices=NATTEN_FMHA_BACKENDS,
        help="Backend / kernel for cross-attention (additional KV) and fast-path self attention in NATTEN. "
        "Choices: "
        f"{', '.join(NATTEN_FMHA_BACKENDS)}.",
    )

    parser.add_argument(
        "--q-tile",
        type=int,
        nargs="*",
        help="Q tile shape in the forward pass kernel (varies between different backends). Try "
        "--dry-run to see available backends and tile shapes for your use case.",
    )

    parser.add_argument(
        "--kv-tile",
        type=int,
        nargs="*",
        help="KV tile shape in the forward pass kernel (varies between different backends). Try "
        "--dry-run to see available backends and tile shapes for your use case.",
    )

    parser.add_argument(
        "--backward-q-tile",
        type=int,
        nargs="*",
        help="Q tile shape in the backward pass kernel (only respected by `cutlass-fna` backend). "
        "Try --dry-run to see available backends and tile shapes for your use case.",
    )

    parser.add_argument(
        "--backward-kv-tile",
        type=int,
        nargs="*",
        help="KV tile shape in the backward pass kernel (only respected by `cutlass-fna` backend). "
        "Try --dry-run to see available backends and tile shapes for your use case.",
    )

    parser.add_argument(
        "--schedule",
        type=str,
        default=None,
        choices=SCHEDULE_MAP.keys(),
        help="Kernel schedule (hopper-fna and hopper-fmha only). "
        "Choices: non (non-persistent), coop (cooperative), pp (ping-pong).",
    )

    parser.add_argument(
        "--persistent",
        action="store_true",
        help="Use persistent scheduling in the kernel (blackwell-fna and blackwell-fmha only).",
    )

    parser.add_argument(
        "--compile",
        action="store_true",
        help="Compile Flex Attention block sparse mask and kernel (flex backend only).",
    )

    parser.add_argument(
        "--warmup-steps",
        type=int,
        default=10,
        help="Number of warmup steps. Defaults to 10.",
    )

    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Display valid forward and backward pass configurations for this use case and exit.",
    )

    parser.add_argument(
        "--max-configs",
        type=int,
        default=10,
        help="Maximum number of tile configurations to display in dry run. Defaults to 10. If set "
        "to 0 shows all valid configurations.",
    )

    parser.add_argument(
        "--optimize",
        action="store_true",
        help="Find the best configuration (and backend if unspecified) for your use case, by "
        "profiling all available choices, and selecting the fastest one.",
    )

    parser.add_argument(
        "--optimize-warmup-steps",
        type=int,
        default=5,
        help="Number of warmup steps for optimizer. Defaults to 5.",
    )

    args = parser.parse_args()

    args.dim_value = args.dim_value or args.dim

    args.input_size = check_input_size(args.input_size)
    args.window_size = check_window_size(
        input_size=args.input_size, window_size=args.window_size
    )
    args.dilation = check_dilation(
        input_size=args.input_size, window_size=args.window_size, dilation=args.dilation
    )
    args.stride = check_stride(
        input_size=args.input_size,
        window_size=args.window_size,
        dilation=args.dilation,
        stride=args.stride,
    )
    args.causal = check_causal(input_size=args.input_size, causal=args.causal)

    return args


def profile(
    batch_size: int,
    heads: int,
    input_size,
    dim: int,
    dim_value: int,
    window_size: DimensionType,
    stride: DimensionType,
    dilation: DimensionType,
    dtype: str,
    causal: CausalArgType,
    warmup_steps: int,
    backend: str,
    fmha_backend: str,
    backprop: bool,
    add_kv: int,
    q_tile: Optional[DimensionType],
    kv_tile: Optional[DimensionType],
    backward_q_tile: Optional[DimensionType],
    backward_kv_tile: Optional[DimensionType],
    persistent: bool,
    schedule: Optional[str],
    compile: bool,
    dry_run: bool,
    max_configs: int,
    optimize: bool,
    optimize_warmup_steps: int,
):
    if dry_run and optimize:
        raise ValueError("Dry run and optimize can't be run together.")

    if (q_tile is not None) ^ (kv_tile is not None):
        raise ValueError(
            "Q tile and KV tile must be set together, got " f"{q_tile=}, {kv_tile=}."
        )

    if (backward_q_tile is not None) ^ (backward_kv_tile is not None):
        raise ValueError(
            "Backward Q tile and KV tile must be set together, got "
            f"{backward_q_tile=}, {backward_kv_tile=}."
        )

    q_tile_size = None if q_tile is None else math.prod(q_tile)
    kv_tile_size = None if kv_tile is None else math.prod(kv_tile)
    backward_q_tile_size = (
        None if backward_q_tile is None else math.prod(backward_q_tile)
    )
    backward_kv_tile_size = (
        None if backward_kv_tile is None else math.prod(backward_kv_tile)
    )

    if dtype == "fp16":
        torch_dtype = torch.float16
    elif dtype == "bf16":
        torch_dtype = torch.bfloat16
    elif dtype == "fp32":
        torch_dtype = torch.float32
    else:
        raise NotImplementedError(
            f"Data type {dtype} is not supported. Choices are " "fp16, bf16, fp32."
        )

    if compile:
        allow_flex_compile(True, True)

    use_kv_parallelism_in_fused_na(True)
    set_memory_usage_preference("unrestricted")

    problem = generate_problem(
        batch_size=batch_size,
        heads=heads,
        input_size=input_size,
        dim=dim,
        dim_value=dim_value,
        window_size=window_size,
        stride=stride,
        dilation=dilation,
        dtype=torch_dtype,
        is_causal=causal,
        additional_kv_length=add_kv,
    )

    if dry_run:
        dry_run_fn(
            problem,
            backend=backend,
            fmha_backend=fmha_backend,
            backprop=backprop,
            torch_compile=compile,
            max_configs=max_configs,
        )
        return

    if optimize:
        (
            opt_backend,
            opt_fmha_backend,
            opt_q_tile,
            opt_kv_tile,
            opt_backward_q_tile,
            opt_backward_kv_tile,
            opt_q_tile_size,
            opt_kv_tile_size,
            opt_backward_q_tile_size,
            opt_backward_kv_tile_size,
            opt_schedule,
        ) = optimize_fn(
            problem=problem,
            backend=backend,
            fmha_backend=fmha_backend,
            q_tile=q_tile,
            kv_tile=kv_tile,
            backward_q_tile=backward_q_tile,
            backward_kv_tile=backward_kv_tile,
            backprop=backprop,
            schedule=schedule,
            persistent=persistent,
            torch_compile=compile,
            warmup_steps=optimize_warmup_steps,
            q_tile_size=q_tile_size,
            kv_tile_size=kv_tile_size,
            backward_q_tile_size=backward_q_tile_size,
            backward_kv_tile_size=backward_kv_tile_size,
        )

        logged_ops = do_profile(
            problem=problem,
            backend=opt_backend,
            fmha_backend=opt_fmha_backend,
            q_tile_shape=opt_q_tile,
            kv_tile_shape=opt_kv_tile,
            backward_q_tile_shape=opt_backward_q_tile,
            backward_kv_tile_shape=opt_backward_kv_tile,
            persistent=persistent,
            torch_compile=compile,
            kernel_schedule=opt_schedule,
            warmup_steps=warmup_steps,
            backprop=backprop,
            q_tile_size=opt_q_tile_size,
            kv_tile_size=opt_kv_tile_size,
            backward_q_tile_size=opt_backward_q_tile_size,
            backward_kv_tile_size=opt_backward_kv_tile_size,
        )

    else:

        logged_ops = do_profile(
            problem=problem,
            backend=backend,
            fmha_backend=fmha_backend,
            q_tile_shape=q_tile,
            kv_tile_shape=kv_tile,
            backward_q_tile_shape=backward_q_tile,
            backward_kv_tile_shape=backward_kv_tile,
            persistent=persistent,
            torch_compile=compile,
            kernel_schedule=schedule,
            warmup_steps=warmup_steps,
            backprop=backprop,
            q_tile_size=q_tile_size,
            kv_tile_size=kv_tile_size,
            backward_q_tile_size=backward_q_tile_size,
            backward_kv_tile_size=backward_kv_tile_size,
        )

    title = "Profiler results"
    headers = [
        "Framework",
        "Kernel category",
        "Arch",
        "Operation",
        "# calls",
        "Runtime",
    ]
    assert all(isinstance(x, Result) for x in logged_ops), f"{logged_ops}"
    values = [
        [
            x.framework,
            x.kernel_type,
            x.tag,
            x.op_name_formatted,
            str(x.num_calls),
            x.time_str,
        ]
        for x in logged_ops
    ]
    total = sum(logged_ops)
    assert isinstance(total, Result)
    values.append(["", "", "", "Total", "", total.time_str])
    print_table(title, headers, values, has_footer=True)


if __name__ == "__main__":
    args = get_args()
    profile(**vars(args))



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from typing import List, Optional

import torch  # noqa: F401
from torch import Tensor

from ....types import (
    CutlassBlackwellFmhaBackwardConfigType,
    CutlassBlackwellFmhaForwardConfigType,
    CutlassBlackwellFnaBackwardConfigType,
    CutlassBlackwellFnaForwardConfigType,
    DimensionType,
)
from ....utils.checks import check_tile_shape
from ....utils.device import get_device_cc


# The current CUTLASS FMHA forward kernel can only do Q tile size 256, KV tile size 128.
# This limits 1D tile shapes to just the one, but for 2-D and 3-D we can have many more shapes,
# only some of which we compile. Adding new ones requires adding them to autogen, regenerating
# the instantiations, and recompiling libnatten. Unlike CUTLASS 2.X FNA, multi-dim tile shapes are
# static in Blackwell FNA, and not dynamic.

BLACKWELL_FORWARD_TILE_SHAPES = {
    1: [
        ((256,), (128,)),
    ],
    2: [
        ((16, 16), (16, 8)),
        ((16, 16), (8, 16)),
        ((8, 32), (8, 16)),
        ((8, 32), (4, 32)),
    ],
    3: [
        ((8, 4, 8), (4, 4, 8)),
        ((8, 4, 8), (2, 8, 8)),
        ((2, 8, 16), (4, 4, 8)),
        ((2, 8, 16), (2, 8, 8)),
        ((4, 4, 16), (2, 4, 16)),
        ((2, 16, 8), (2, 8, 8)),
        ((4, 8, 8), (2, 8, 8)),
    ],
}

BLACKWELL_BACKWARD_TILE_SHAPES = {
    1: [
        ((128,), (128,)),
    ],
    2: [
        ((16, 8), (16, 8)),
        ((16, 8), (8, 16)),
        ((8, 16), (16, 8)),
        ((8, 16), (8, 16)),
    ],
    3: [
        ((4, 4, 8), (4, 4, 8)),
        ((4, 4, 8), (2, 8, 8)),
        ((1, 8, 16), (4, 4, 8)),
        ((2, 8, 8), (4, 4, 8)),
        ((1, 8, 16), (2, 8, 8)),
        ((2, 4, 16), (2, 4, 16)),
        ((4, 2, 16), (2, 4, 16)),
        ((4, 4, 8), (2, 4, 16)),
        ((2, 8, 8), (2, 8, 8)),
    ],
}


def _get_default_tile_shapes_forward(
    na_dim: int,
) -> CutlassBlackwellFnaForwardConfigType:
    assert na_dim in [1, 2, 3]

    if na_dim == 1:
        return ((256,), (128,))
    if na_dim == 2:
        return ((16, 16), (16, 8))
    if na_dim == 3:
        return ((8, 4, 8), (4, 4, 8))

    raise NotImplementedError()


def get_all_forward_configs(
    input_tensor: Tensor,
) -> List[CutlassBlackwellFnaForwardConfigType]:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim

    device_cc = get_device_cc(input_tensor.device)
    if device_cc != 100:
        return []

    return BLACKWELL_FORWARD_TILE_SHAPES[na_dim]  # type: ignore


def get_all_backward_configs(
    input_tensor: Tensor,
) -> List[CutlassBlackwellFnaBackwardConfigType]:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim

    device_cc = get_device_cc(input_tensor.device)
    if device_cc != 100:
        return []

    return BLACKWELL_BACKWARD_TILE_SHAPES[na_dim]  # type: ignore


# For FMHA
def get_all_fmha_forward_configs(
    input_tensor: Tensor,
) -> List[CutlassBlackwellFmhaForwardConfigType]:
    if input_tensor.dim() != 4:
        raise ValueError("Only 4-D tensors are supported in FMHA.")

    tile_shapes = get_all_forward_configs(input_tensor)
    assert all(len(q_t) == len(kv_t) == 1 for q_t, kv_t in tile_shapes)

    tile_sizes = [(q_t[0], kv_t[0]) for q_t, kv_t in tile_shapes]

    return tile_sizes


def get_all_fmha_backward_configs(
    input_tensor: Tensor,
) -> List[CutlassBlackwellFmhaBackwardConfigType]:
    if input_tensor.dim() != 4:
        raise ValueError("Only 4-D tensors are supported in FMHA.")

    tile_shapes = get_all_backward_configs(input_tensor)
    assert all(len(q_t) == len(kv_t) == 1 for q_t, kv_t in tile_shapes)

    tile_sizes = [(q_t[0], kv_t[0]) for q_t, kv_t in tile_shapes]

    return tile_sizes


def get_default_forward_tile_shapes(
    input_tensor: Tensor,
) -> CutlassBlackwellFnaForwardConfigType:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim

    return _get_default_tile_shapes_forward(na_dim)


def get_default_forward_tile_sizes(
    input_tensor: Tensor,
) -> CutlassBlackwellFmhaForwardConfigType:
    if input_tensor.dim() != 4:
        raise ValueError("Only 4-D tensors are supported in FMHA.")

    q_t, kv_t = get_default_forward_tile_shapes(input_tensor)
    assert len(q_t) == len(kv_t) == 1

    return (q_t[0], kv_t[0])


def get_default_backward_tile_shapes(
    input_tensor: Tensor,
) -> CutlassBlackwellFnaBackwardConfigType:
    all_configs = get_all_backward_configs(input_tensor)

    if len(all_configs) < 1:
        device_cc = get_device_cc(input_tensor.device)
        raise ValueError(
            "No configs exist for this use case; Blackwell FMHA/FNA does not support it: "
            f"{input_tensor.shape=}, {input_tensor.dtype=}, {device_cc=}."
        )

    return all_configs[0]


def get_default_backward_tile_sizes(
    input_tensor: Tensor,
) -> CutlassBlackwellFmhaBackwardConfigType:
    if input_tensor.dim() != 4:
        raise ValueError("Only 4-D tensors are supported in FMHA.")

    q_t, kv_t = get_default_backward_tile_shapes(input_tensor)
    assert len(q_t) == len(kv_t) == 1

    return (q_t[0], kv_t[0])


def check_cutlass_blackwell_fna_forward_config(
    input_tensor: Tensor,
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
) -> CutlassBlackwellFnaForwardConfigType:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim

    if (q_tile_shape is None) ^ (kv_tile_shape is None):
        raise ValueError(
            "Please specify both q_tile_shape and kv_tile_shape, or neither one. "
            f"Got {q_tile_shape=}, {kv_tile_shape=}."
        )

    if q_tile_shape is None and kv_tile_shape is None:
        return get_default_forward_tile_shapes(input_tensor=input_tensor)

    q_tile_shape = check_tile_shape(q_tile_shape)
    kv_tile_shape = check_tile_shape(kv_tile_shape)

    tile_shapes = get_all_forward_configs(input_tensor=input_tensor)

    for q_t, kv_t in tile_shapes:
        if q_t == q_tile_shape and kv_t == kv_tile_shape:
            return (q_t, kv_t)  # type: ignore

    # Fail and make suggestions
    MAX_EXAMPLES = 3
    examples = ""
    for i, (q_t, kv_t) in enumerate(tile_shapes):
        examples += f"\n  q_tile_shape={q_t}, kv_tile_shape={kv_t}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for CUTLASS Blackwell FNA-{na_dim}D. "
        f"Q tile shape {q_tile_shape} and KV tile shape {kv_tile_shape} "
        f"are not among the {len(tile_shapes)} configurations implementable "
        f"with CUTLASS Blackwell FNA. "
        "Try selecting a combination from: \n"
        "  natten.get_configs_for_blackwell_fna(q, k, v)"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )


def check_cutlass_blackwell_fmha_forward_config(
    input_tensor: Tensor,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
) -> CutlassBlackwellFmhaForwardConfigType:
    assert input_tensor.dim() == 4

    if (q_tile_size is None) ^ (kv_tile_size is None):
        raise ValueError(
            "Please specify both q_tile_size and kv_tile_size, or neither one. "
            f"Got {q_tile_size=}, {kv_tile_size=}."
        )

    if q_tile_size is None and kv_tile_size is None:
        q_tile_size, kv_tile_size = get_default_forward_tile_sizes(
            input_tensor=input_tensor
        )
        return (q_tile_size, kv_tile_size)

    tile_sizes = get_all_fmha_forward_configs(input_tensor=input_tensor)

    for q_t, kv_t in tile_sizes:
        if q_t == q_tile_size and kv_t == kv_tile_size:
            return (q_t, kv_t)

    # Fail and make suggestions
    MAX_EXAMPLES = 3
    examples = ""
    for i, (q_t, kv_t) in enumerate(tile_sizes):
        examples += f"\n  q_tile_size={q_t}, kv_tile_size={kv_t}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for CUTLASS Blackwell FMHA. "
        f"Q tile size {q_tile_size} and KV tile size {kv_tile_size} "
        f"are not among the {len(tile_sizes)} configurations implementable "
        f"with CUTLASS Blackwell FMHA. "
        "Try selecting a combination from: \n"
        "  natten.get_configs_for_blackwell_fmha(q, k, v)"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )


def check_cutlass_blackwell_fna_backward_config(
    input_tensor: Tensor,
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
) -> CutlassBlackwellFnaBackwardConfigType:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim

    if (q_tile_shape is None) ^ (kv_tile_shape is None):
        raise ValueError(
            "Please specify both q_tile_shape and kv_tile_shape, or neither one. "
            f"Got {q_tile_shape=}, {kv_tile_shape=}."
        )

    if q_tile_shape is None and kv_tile_shape is None:
        return get_default_backward_tile_shapes(input_tensor=input_tensor)

    q_tile_shape = check_tile_shape(q_tile_shape)
    kv_tile_shape = check_tile_shape(kv_tile_shape)

    tile_shapes = get_all_backward_configs(input_tensor=input_tensor)

    for q_t, kv_t in tile_shapes:
        if q_t == q_tile_shape and kv_t == kv_tile_shape:
            return (q_t, kv_t)  # type: ignore

    # Fail and make suggestions
    MAX_EXAMPLES = 3
    examples = ""
    for i, (q_t, kv_t) in enumerate(tile_shapes):
        examples += f"\n  q_tile_shape={q_t}, kv_tile_shape={kv_t}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for CUTLASS Blackwell FNA-{na_dim}D Backward. "
        f"Q tile shape {q_tile_shape} and KV tile shape {kv_tile_shape} "
        f"are not among the {len(tile_shapes)} configurations implementable "
        f"with CUTLASS Blackwell FNA Backward. "
        "Try selecting a combination from: \n"
        "  natten.get_bwd_configs_for_blackwell_fna(q, k, v)"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )


def check_cutlass_blackwell_fmha_backward_config(
    input_tensor: Tensor,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
) -> CutlassBlackwellFmhaBackwardConfigType:
    assert input_tensor.dim() == 4

    if (q_tile_size is None) ^ (kv_tile_size is None):
        raise ValueError(
            "Please specify both q_tile_size and kv_tile_size, or neither one. "
            f"Got {q_tile_size=}, {kv_tile_size=}."
        )

    if q_tile_size is None and kv_tile_size is None:
        q_tile_size, kv_tile_size = get_default_backward_tile_sizes(
            input_tensor=input_tensor
        )
        return (q_tile_size, kv_tile_size)

    tile_sizes = get_all_fmha_backward_configs(input_tensor=input_tensor)

    for q_t, kv_t in tile_sizes:
        if q_t == q_tile_size and kv_t == kv_tile_size:
            return (q_t, kv_t)

    # Fail and make suggestions
    MAX_EXAMPLES = 3
    examples = ""
    for i, (q_t, kv_t) in enumerate(tile_sizes):
        examples += f"\n  q_tile_size={q_t}, kv_tile_size={kv_t}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for CUTLASS Blackwell FMHA Backward. "
        f"Q tile size {q_tile_size} and KV tile size {kv_tile_size} "
        f"are not among the {len(tile_sizes)} configurations implementable "
        f"with CUTLASS Blackwell FMHA Backward. "
        "Try selecting a combination from: \n"
        "  natten.get_bwd_configs_for_blackwell_fmha(q, k, v)"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################
import functools
from typing import Optional, Tuple, Union

import torch
from torch import Tensor

from torch.amp import custom_bwd, custom_fwd
from torch.autograd import Function

amp_fwd = functools.partial(custom_fwd, device_type="cuda")
amp_bwd = functools.partial(custom_bwd, device_type="cuda")

from .._libnatten import (
    reference_na1d_backward,
    reference_na1d_forward,
    reference_na2d_backward,
    reference_na2d_forward,
    reference_na3d_backward,
    reference_na3d_forward,
)
from ..types import (
    CausalArg1DTypeOrDed,
    CausalArg2DTypeOrDed,
    CausalArg3DTypeOrDed,
    CausalArgType,
    CausalArgTypeOrDed,
    Dimension1DTypeOrDed,
    Dimension2DTypeOrDed,
    Dimension3DTypeOrDed,
    DimensionType,
    DimensionTypeOrDed,
    NoneType,
)
from ..utils import log
from ..utils.checks import (
    additional_kv_tensor_checks,
    check_all_args,
    check_args_against_input,
    na_tensor_checks,
)

logger = log.get_logger(__name__)


def make_reference_fna_autograd_fn(na_dim):
    assert na_dim in [1, 2, 3]

    FORWARD_OPS = {
        1: reference_na1d_forward,
        2: reference_na2d_forward,
        3: reference_na3d_forward,
    }

    BACKWARD_OPS = {
        1: reference_na1d_backward,
        2: reference_na2d_backward,
        3: reference_na3d_backward,
    }

    class ReferenceFnaGenericAutogradFn(Function):
        @staticmethod
        @amp_fwd
        def forward(
            ctx,
            query: Tensor,
            key: Tensor,
            value: Tensor,
            kernel_size: DimensionType,
            stride: DimensionType,
            dilation: DimensionType,
            is_causal: CausalArgType,
            scale: float,
            qkv_shape: DimensionType,
            num_extra_kv: int,
        ) -> Tuple[Tensor, Tensor]:
            kernel_size, stride, dilation, is_causal = check_all_args(
                na_dim, kernel_size, stride, dilation, is_causal
            )

            query = query.contiguous()
            key = key.contiguous()
            value = value.contiguous()

            assert query.dim() == value.dim() == 4
            assert query.shape[0] == value.shape[0]
            assert query.shape[-2] == value.shape[-2]

            output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
            output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

            logsumexp = torch.empty(
                query.shape[:-1], dtype=torch.float32, device=query.device
            )

            FORWARD_OPS[na_dim](
                output,
                query,
                key,
                value,
                logsumexp,
                kernel_size,
                stride,
                dilation,
                is_causal,
                scale,
                qkv_shape,
                num_extra_kv,
            )

            ctx.save_for_backward(query, key, value, logsumexp, output)
            ctx.kernel_size = kernel_size
            ctx.stride = stride
            ctx.dilation = dilation
            ctx.is_causal = is_causal
            ctx.scale = scale
            ctx.qkv_shape = qkv_shape
            ctx.num_extra_kv = num_extra_kv

            return output, logsumexp

        @staticmethod
        @amp_bwd
        def backward(ctx, grad_out: Tensor, grad_lse: Tensor) -> Tuple[
            Tensor,
            Tensor,
            Tensor,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
        ]:
            query, key, value, logsumexp, output = ctx.saved_tensors
            d_output = grad_out.contiguous()
            d_query = torch.empty_like(query)
            d_key = torch.empty_like(key)
            d_value = torch.empty_like(value)

            BACKWARD_OPS[na_dim](
                d_query,
                d_key,
                d_value,
                query,
                key,
                value,
                output,
                d_output,
                logsumexp,
                ctx.kernel_size,
                ctx.stride,
                ctx.dilation,
                ctx.is_causal,
                ctx.scale,
                ctx.qkv_shape,
                ctx.num_extra_kv,
            )

            return d_query, d_key, d_value, None, None, None, None, None, None, None

    return ReferenceFnaGenericAutogradFn


ReferenceFna1DAutogradFn = make_reference_fna_autograd_fn(1)
ReferenceFna2DAutogradFn = make_reference_fna_autograd_fn(2)
ReferenceFna3DAutogradFn = make_reference_fna_autograd_fn(3)


ReferenceFnaAutogradFns = {
    1: ReferenceFna1DAutogradFn,
    2: ReferenceFna2DAutogradFn,
    3: ReferenceFna3DAutogradFn,
}


def reference_fna_generic(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: DimensionTypeOrDed,
    stride: DimensionTypeOrDed = 1,
    dilation: DimensionTypeOrDed = 1,
    is_causal: Optional[CausalArgTypeOrDed] = False,
    scale: Optional[float] = None,
    additional_keys: Optional[Tensor] = None,
    additional_values: Optional[Tensor] = None,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:

    na_tensor_checks(query, key, value, must_match_head_dims=False)
    additional_kv_tensor_checks(
        query,
        key,
        value,
        additional_keys,
        additional_values,
        must_match_head_dims=False,
    )

    na_dim = query.dim() - 3  # batch, heads, head_dim

    assert na_dim in [1, 2, 3]

    kernel_size, stride, dilation, is_causal = check_all_args(
        na_dim, kernel_size, stride, dilation, is_causal
    )

    check_args_against_input(
        query,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
    )

    scale = scale or query.shape[-1] ** -0.5

    qkv_shape = query.shape[1 : 1 + na_dim]

    query = query.flatten(1, na_dim)
    key = key.flatten(1, na_dim)
    value = value.flatten(1, na_dim)

    num_extra_kv = 0
    if additional_keys is not None and additional_values is not None:
        num_extra_kv = additional_keys.shape[1]
        key = torch.cat([key, additional_keys], dim=1)
        value = torch.cat([value, additional_values], dim=1)

    output, lse = ReferenceFnaAutogradFns[na_dim].apply(
        query,
        key,
        value,
        kernel_size,
        stride,
        dilation,
        is_causal,
        scale,
        qkv_shape,
        num_extra_kv,
    )
    output = output.reshape(
        query.shape[0], *qkv_shape, query.shape[-2], value.shape[-1]
    )
    lse = lse.reshape(query.shape[0], *qkv_shape, query.shape[-2])

    if return_lse:
        return output, lse

    return output


def na1d_reference(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension1DTypeOrDed,
    stride: Dimension1DTypeOrDed = 1,
    dilation: Dimension1DTypeOrDed = 1,
    is_causal: Optional[CausalArg1DTypeOrDed] = False,
    scale: Optional[float] = None,
    additional_keys: Optional[Tensor] = None,
    additional_values: Optional[Tensor] = None,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    return reference_fna_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        additional_keys=additional_keys,
        additional_values=additional_values,
        return_lse=return_lse,
    )


def na2d_reference(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension2DTypeOrDed,
    stride: Dimension2DTypeOrDed = 1,
    dilation: Dimension2DTypeOrDed = 1,
    is_causal: Optional[CausalArg2DTypeOrDed] = False,
    scale: Optional[float] = None,
    additional_keys: Optional[Tensor] = None,
    additional_values: Optional[Tensor] = None,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    return reference_fna_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        additional_keys=additional_keys,
        additional_values=additional_values,
        return_lse=return_lse,
    )


def na3d_reference(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension3DTypeOrDed,
    stride: Dimension3DTypeOrDed = 1,
    dilation: Dimension3DTypeOrDed = 1,
    is_causal: Optional[CausalArg3DTypeOrDed] = False,
    scale: Optional[float] = None,
    additional_keys: Optional[Tensor] = None,
    additional_values: Optional[Tensor] = None,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    return reference_fna_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        additional_keys=additional_keys,
        additional_values=additional_values,
        return_lse=return_lse,
    )



#!/usr/bin/env python
#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import multiprocessing
import os
import shutil
import subprocess
import sys
import tempfile
from os import path
from pathlib import Path
from typing import Any, List

import torch
from setuptools import Extension, setup  # type: ignore
from setuptools.command.build_ext import build_ext  # type: ignore
from torch.utils.cpp_extension import LIB_EXT

IS_WINDOWS = sys.platform == "win32"
IS_LINUX = sys.platform.startswith("linux")
IS_LIBTORCH_BUILT_WITH_CXX11_ABI = torch._C._GLIBCXX_USE_CXX11_ABI

this_directory = Path(__file__).parent
try:
    long_description = (this_directory / "docs/README_pypi.md").read_text()
except:
    long_description = "Neighborhood Attention Extension."

torch_ver = [int(x) for x in torch.__version__.split(".")[:2]]

assert torch_ver >= [2, 5], "NATTEN only supports PyTorch >= 2.5"

CUDA_ARCH = os.getenv("NATTEN_CUDA_ARCH", "")
HAS_CUDA_ARCH = CUDA_ARCH != ""

NATTEN_IS_BUILDING_DIST = bool(os.getenv("NATTEN_IS_BUILDING_DIST", "0") == "1")

VERBOSE = bool(os.getenv("NATTEN_VERBOSE", "0") == "1")

tmp_dir = tempfile.TemporaryDirectory()
NATTEN_BUILD_DIR = os.getenv("NATTEN_BUILD_DIR", tmp_dir.name)
if not os.path.isdir(NATTEN_BUILD_DIR):
    NATTEN_BUILD_DIR = tmp_dir.name

DEFAULT_N_WORKERS = max(1, (multiprocessing.cpu_count() // 4))
try:
    N_WORKERS = int(os.getenv("NATTEN_N_WORKERS", DEFAULT_N_WORKERS))
except:
    N_WORKERS = int(DEFAULT_N_WORKERS)

if not HAS_CUDA_ARCH:
    HAS_CUDA = torch.cuda.is_available()

    if HAS_CUDA:
        cuda_device = torch.cuda.get_device_properties(torch.cuda.current_device())
        sm = cuda_device.major + cuda_device.minor * 0.1
        CUDA_ARCH = f"{sm}"
        print(
            "`NATTEN_CUDA_ARCH` not set, but detected CUDA driver with PyTorch. "
            f"Building for {CUDA_ARCH=}."
        )

        assert torch.version.cuda is not None
        TORCH_CUDA_VERSION = [x for x in torch.version.cuda.split(".")[:2]]
        CUDA_TAG = "".join([x for x in TORCH_CUDA_VERSION])
        CUDA_VERSION = [int(x) for x in TORCH_CUDA_VERSION]

        assert CUDA_VERSION >= [12, 0], "NATTEN only supports CUDA 12.0 and above."
        if CUDA_VERSION >= [12, 0] and IS_WINDOWS:
            print(
                "WARNING: Torch cmake will likely fail on Windows with CUDA 12.X. "
                "Please refer to NATTEN documentation to read more about the issue "
                "and how to get around it until the issue is fixed in torch."
            )

        print(f"PyTorch was built with CUDA Toolkit {CUDA_TAG}")
        print(f"Building NATTEN for the following architecture(s): {CUDA_ARCH}")

        print(f"Number of workers: {N_WORKERS}")

    else:
        print(
            "Building WITHOUT libnatten. `NATTEN_CUDA_ARCH` is not set, and did not detect CUDA "
            "driver with PyTorch."
        )


BUILD_WITH_CUDA = CUDA_ARCH != ""


def get_version() -> str:
    init_py_path = path.join(
        path.abspath(path.dirname(__file__)), "src/natten", "version.py"
    )
    init_py = open(init_py_path, "r").readlines()
    version_line = [ln.strip() for ln in init_py if ln.startswith("__version__")][0]
    version = version_line.split("=")[-1].strip().strip("'\"")
    if not NATTEN_IS_BUILDING_DIST:
        return version
    PYTORCH_VERSION = "".join(torch.__version__.split("+")[0].split("."))

    if BUILD_WITH_CUDA:
        if torch.version.cuda is None:
            raise ValueError(
                "Attempted to build NATTEN with libnatten, which requires PyTorch be "
                f"built with CUDA, but {torch.version.cuda=}."
            )

        TORCH_CUDA_VERSION = [x for x in torch.version.cuda.split(".")[:2]]
        CUDA_TAG = "".join([x for x in TORCH_CUDA_VERSION])
        CU = f"cu{CUDA_TAG}"
        version = f"{version}+torch{PYTORCH_VERSION}{CU}"

    return version


def _check_cuda_arch(arch: Any) -> int:
    try:
        arch = float(arch)
        arch = arch * 10  # 8.6 => 86
        assert (
            arch >= 30
        ), f"Only SM30 and above are supported at this time, got {arch}."
        return int(arch)
    except ValueError:
        raise ValueError(f"Invalid architecture list {arch}.")


def get_cuda_arch_list(cuda_arch_list: str) -> List[int]:
    sep = ";"  # expects semicolon separated list
    if sep not in cuda_arch_list:
        # Probably a single arch
        return [_check_cuda_arch(cuda_arch_list)]
    arch_list = cuda_arch_list.split(sep)
    output_arch_list = []
    for arch in arch_list:
        output_arch_list.append(_check_cuda_arch(arch))
    return output_arch_list


def arch_list_to_cmake_tags(arch_list: List[int]) -> str:
    return (
        "-real;".join([str(x) if x not in [90, 100] else f"{x}a" for x in arch_list])
        + "-real"
    )


class BuildExtension(build_ext):
    def build_extension(self, ext):
        if BUILD_WITH_CUDA:
            # Hack so that we can build somewhere other than /tmp in development mode.
            # Also because we want CMake to build everything elsewhere, otherwise pypi will package
            # build files.
            build_dir = self.build_lib if NATTEN_BUILD_DIR is None else NATTEN_BUILD_DIR

            this_dir = path.dirname(path.abspath(__file__))
            cmake_lists_dir = path.join(this_dir, "csrc")
            try:
                subprocess.check_output(["cmake", "--version"])
            except OSError:
                raise RuntimeError("Cannot find CMake executable")

            output_binary_name = self.get_ext_filename(
                ext.name
            )  # i.e. libnatten.cpython-VERSION-ARCH-OS.so
            output_so_name = output_binary_name.replace(LIB_EXT, "")

            # extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

            max_sm = 0
            cuda_arch_list = []
            cuda_arch_list_str = ""

            cuda_arch_list = get_cuda_arch_list(CUDA_ARCH)  # Expects xx -- i.e. 86
            cuda_arch_list_str = arch_list_to_cmake_tags(cuda_arch_list)
            max_sm = max(cuda_arch_list)

            print(
                f"Building NATTEN for the following archs: {cuda_arch_list} (max: {max_sm})"
            )
            print(f"Building with {N_WORKERS} workers.")
            print(f"Build directory: {build_dir}")
            print(f"{IS_LIBTORCH_BUILT_WITH_CXX11_ABI=}")

            cmake_args = [
                f"-DPYTHON_PATH={sys.executable}",
                f"-DOUTPUT_FILE_NAME={output_so_name}",
                f"-DNATTEN_CUDA_ARCH_LIST={cuda_arch_list_str}",
                f"-DNATTEN_IS_WINDOWS={int(IS_WINDOWS)}",
                f"-DIS_LIBTORCH_BUILT_WITH_CXX11_ABI={int(IS_LIBTORCH_BUILT_WITH_CXX11_ABI)}",
            ]

            if max_sm < 50:
                raise RuntimeError(
                    "NATTEN's CUDA backend only supports SM50 and above, "
                    f"saw SM{max_sm} in {CUDA_ARCH}."
                )

            if 90 in cuda_arch_list:
                cmake_args.append("-DNATTEN_WITH_HOPPER_FNA=1")

            if 100 in cuda_arch_list:
                cmake_args.append("-DNATTEN_WITH_BLACKWELL_FNA=1")

            if IS_WINDOWS:
                python_path = sys.executable
                assert (
                    "python.exe" in python_path
                ), f"Expected the python executable path to end with python.exe, got {python_path}"
                python_lib_dir = python_path.replace("python.exe", "libs").strip()
                cmake_args.append(f"-DPY_LIB_DIR={python_lib_dir}")
                cmake_args.append("-G Ninja")
                cmake_args.append("-DCMAKE_BUILD_TYPE=Release")

            if not os.path.exists(build_dir):
                os.makedirs(build_dir)

            if not os.path.exists(self.build_lib):
                os.makedirs(self.build_lib)

            so_dir_local = os.path.join(build_dir, os.path.dirname(output_binary_name))
            so_path_local = f"{build_dir}/{output_binary_name}"
            if not os.path.exists(so_dir_local):
                os.makedirs(so_dir_local)

            so_dir = os.path.join(self.build_lib, os.path.dirname(output_binary_name))
            so_path_final = f"{self.build_lib}/{output_binary_name}"
            if not os.path.exists(so_dir):
                os.makedirs(so_dir)

            # Config and build the extension
            subprocess.check_call(
                ["cmake", cmake_lists_dir] + cmake_args, cwd=build_dir
            )
            cmake_build_args = [
                "--build",
                build_dir,
                "-j",
                str(N_WORKERS),
            ]
            if VERBOSE:
                cmake_build_args.append("--verbose")
            subprocess.check_call(["cmake", *cmake_build_args])

            assert os.path.isfile(
                so_path_local
            ), f"Expected libnatten binary in {so_path_local}."
            if build_dir != self.build_lib:
                shutil.copy(so_path_local, so_path_final)
            assert os.path.isfile(so_path_final)

            # Clean up cmake files when building dist package;
            # otherwise they will get packed into the wheel.
            if NATTEN_IS_BUILDING_DIST:
                for file in os.listdir(build_dir):
                    fn = os.path.join(build_dir, file)
                    if os.path.isfile(fn):
                        os.remove(fn)
                    elif file != "natten":
                        shutil.rmtree(fn)
        else:
            # Libnatten is CUDA only now.
            pass


setup(
    name="natten",
    version=get_version(),
    author="Ali Hassani",
    url="https://natten.org",
    description="Neighborhood Attention Extension.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_dir={"": "src"},
    packages=[
        "natten",
        "natten/profiling_utils",
        "natten/utils",
        "natten/backends",
        "natten/backends/configs",
        "natten/backends/configs/cutlass",
        "natten/backends/configs/cutlass_blackwell",
        "natten/backends/configs/cutlass_hopper",
        "natten/backends/configs/flex",
    ],
    package_data={
        "": ["csrc/**/*"],
    },
    python_requires=">=3.9",
    install_requires=[
        "torch",
    ],
    extras_require={},
    ext_modules=[Extension("natten.libnatten", [])] if BUILD_WITH_CUDA else [],
    cmdclass={"build_ext": BuildExtension} if BUILD_WITH_CUDA else {},
)



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import torch

from .._environment import _IS_CUDA_AVAILABLE, _RUN_EXTENDED_TESTS, HAS_LIBNATTEN

from ..backends.flex import _FLEX_COMPILE_SUPPORTED, _FLEX_SUPPORTED
from .device import get_device_cc, is_cuda


def skip_if_libnatten_is_not_supported():
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            if not _IS_CUDA_AVAILABLE:
                self.skipTest("CUDA is not available.")
            elif not HAS_LIBNATTEN:
                self.skipTest("Libnatten is not available.")
            else:
                return f(self, *args, **kwargs)

        return wrapper

    return decorator


def skip_if_cuda_is_not_supported():
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            if not _IS_CUDA_AVAILABLE:
                self.skipTest("CUDA is not available.")
            else:
                return f(self, *args, **kwargs)

        return wrapper

    return decorator


def skip_if_flex_is_not_supported():
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            if not _FLEX_SUPPORTED or get_device_cc() < 70:
                self.skipTest("Flex backend is not supported.")
            else:
                return f(self, *args, **kwargs)

        return wrapper

    return decorator


def skip_if_flex_compile_is_not_supported():
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            if not _FLEX_COMPILE_SUPPORTED:
                self.skipTest("Flex (compiled) backend is not supported.")
            else:
                return f(self, *args, **kwargs)

        return wrapper

    return decorator


def skip_if_not_running_extended_tests():
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            if not _RUN_EXTENDED_TESTS:
                self.skipTest("Skipping extended test cases.")
            else:
                return f(self, *args, **kwargs)

        return wrapper

    return decorator


def skip_if_hopper_kernels_not_supported():
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            if get_device_cc() != 90:
                self.skipTest("Hopper kernels are only supported on SM90.")
            else:
                return f(self, *args, **kwargs)

        return wrapper

    return decorator


def skip_if_blackwell_kernels_not_supported():
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            if get_device_cc() != 100:
                self.skipTest("Blackwell kernels are only supported on SM100.")
            else:
                return f(self, *args, **kwargs)

        return wrapper

    return decorator


def supports_float16(device: torch.device) -> bool:
    if is_cuda(device):
        device_cc = get_device_cc(device)

        if device_cc < 50:
            return False

        return True

    # TODO:
    return True


def supports_bfloat16(device: torch.device) -> bool:
    if is_cuda(device):
        device_cc = get_device_cc(device)

        if device_cc < 80:
            return False

        return True

    # TODO:
    return False



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import functools
import math
import time
from typing import Callable, Optional, Tuple, Union

import torch
from torch import BoolTensor, IntTensor, Tensor
from torch.nn.attention.flex_attention import (
    BlockMask,
    create_block_mask,
    flex_attention,
)

from ..token_permute import maybe_pad, maybe_unpad, token_permute, token_unpermute
from ..types import (
    CausalArg1DTypeOrDed,
    CausalArg2DTypeOrDed,
    CausalArg3DTypeOrDed,
    CausalArgType,
    CausalArgTypeOrDed,
    Dimension1DType,
    Dimension1DTypeOrDed,
    Dimension2DType,
    Dimension2DTypeOrDed,
    Dimension3DType,
    Dimension3DTypeOrDed,
    DimensionType,
    DimensionTypeOrDed,
)
from ..utils import log
from ..utils.checks import (
    check_all_args,
    check_args_against_input,
    check_input_size_arg,
    fmha_tensor_checks,
    na_tensor_checks,
)

from .configs.checks import (  # noqa: F401
    _FLEX_COMPILE_SUPPORTED,
    _FLEX_SUPPORTED,
    can_run_flex_attention,
)
from .configs.flex import check_flex_fmha_forward_config, check_flex_fna_forward_config

logger = log.get_logger(__name__)


def get_flex_attention_fn(
    torch_compile: bool, torch_compile_args: Optional[dict] = None
) -> Callable:
    if not torch_compile:
        return flex_attention

    additional_args = torch_compile_args or {}
    additional_args["dynamic"] = False

    return torch.compile(flex_attention, **additional_args)


def _run_flex_attn(
    q: Tensor,
    k: Tensor,
    v: Tensor,
    block_mask: BlockMask,
    scale: float,
    torch_compile: bool,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
    torch_compile_args: Optional[dict] = None,
) -> Tuple[Tensor, Tensor]:

    # We may need to override the default flex config.
    # Default ones are not guaranteed to work out of the box across architectures.
    # Some oversubscribe shmem even on the B200!
    torch_compile_args = {}

    # Disable flex decoding path
    kernel_options = {
        "FORCE_USE_FLEX_ATTENTION": True,
    }

    if q_tile_size is not None and torch_compile:
        kv_tile_size = kv_tile_size or q_tile_size

        # Have to auto-tune, otherwise torch will only allow the default config.
        torch_compile_args["mode"] = "max-autotune-no-cudagraphs"

        kernel_options["SPARSE_Q_BLOCK_SIZE"] = q_tile_size  # type: ignore[assignment]
        kernel_options["SPARSE_KV_BLOCK_SIZE"] = kv_tile_size  # type: ignore[assignment]
        kernel_options["BLOCK_M"] = q_tile_size  # type: ignore[assignment]
        kernel_options["BLOCK_N"] = kv_tile_size  # type: ignore[assignment]

    flex_fn = get_flex_attention_fn(
        torch_compile=torch_compile, torch_compile_args=torch_compile_args
    )

    return flex_fn(
        q,
        k,
        v,
        block_mask=block_mask,
        return_lse=True,
        scale=scale,
        kernel_options=kernel_options,
    )


def run_flex_attn(
    q: Tensor,
    k: Tensor,
    v: Tensor,
    block_mask: BlockMask,
    scale: float,
    torch_compile: bool,
    torch_compile_args: Optional[dict] = None,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
) -> Tuple[Tensor, Tensor]:

    if q_tile_size is not None and kv_tile_size is not None:
        return _run_flex_attn(
            q,
            k,
            v,
            block_mask=block_mask,
            scale=scale,
            torch_compile=torch_compile,
            q_tile_size=q_tile_size,
            kv_tile_size=kv_tile_size,
            torch_compile_args=torch_compile_args,
        )

    # Use smallest tile size combo to try and evade shmem oversubscription
    # The defaults just fail very frequently.
    return _run_flex_attn(
        q,
        k,
        v,
        block_mask=block_mask,
        scale=scale,
        torch_compile=torch_compile,
        q_tile_size=64,
        kv_tile_size=64,
        torch_compile_args=torch_compile_args,
    )


def flex_fmha(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    scale: Optional[float] = None,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
    torch_compile: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:

    fmha_tensor_checks(query, key, value, must_match_head_dims=True)

    q_tile_size, kv_tile_size = check_flex_fmha_forward_config(
        input_tensor=query,
        q_tile_size=q_tile_size,
        kv_tile_size=kv_tile_size,
    )

    scale = scale or query.shape[-1] ** -0.5

    assert can_run_flex_attention(
        query, key, value, torch_compile=torch_compile, raise_error=True
    )

    batch_size, seqlen_q, num_heads, head_dim = query.shape
    seqlen_kv = key.shape[1]

    # Flex and torch attention use heads first layout
    query_ = query.reshape(batch_size, seqlen_q, num_heads, head_dim).transpose(1, 2)
    key_ = key.reshape(batch_size, seqlen_kv, num_heads, head_dim).transpose(1, 2)
    value_ = value.reshape(batch_size, seqlen_kv, num_heads, head_dim).transpose(1, 2)

    out_, lse_ = run_flex_attn(
        query_,
        key_,
        value_,
        block_mask=None,  # type: ignore[arg-type]
        scale=scale,
        torch_compile=torch_compile,
        q_tile_size=q_tile_size,
        kv_tile_size=kv_tile_size,
    )

    out = out_.transpose(1, 2).reshape(batch_size, seqlen_q, num_heads, head_dim)
    lse = lse_.transpose(1, 2).reshape(batch_size, seqlen_q, num_heads)

    if return_lse:
        return out, lse

    return out


# TODO: move me elsewhere?
def idx2crd(index, shape) -> tuple:
    rank = len(shape)
    coord = []
    residual = index
    for i in range(rank - 1, -1, -1):
        coord.append(residual % shape[i])
        residual = residual // shape[i]

    # assert residual == 0
    return tuple(coord[::-1])


def get_na_flex_mask(
    device: str,
    na_dim: int,
    num_heads: int,
    qkv_shape: DimensionType,
    kernel_size: DimensionType,
    stride: DimensionType,
    dilation: DimensionType,
    is_causal: CausalArgType,
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
    q_shape: Optional[DimensionType] = None,
    kv_shape: Optional[DimensionType] = None,
    torch_compile: bool = False,
):
    flex_mask_start_time = time.perf_counter()
    do_token_permute = q_tile_shape is not None and kv_tile_shape is not None
    if do_token_permute:
        if q_tile_shape is None or kv_tile_shape is None:
            raise ValueError(
                "Please specify Q and KV tile shapes for multi dimensional tiling. "
                f"Got {q_tile_shape=}, {kv_tile_shape=}."
            )

        if q_shape is None or kv_shape is None:
            raise ValueError(
                "Please specify q_shape and kv_shape for multi dimensional tiling."
            )

        if len(q_tile_shape) != na_dim or len(kv_tile_shape) != na_dim:
            raise ValueError(
                "Q and KV tile shapes must match the number of dimensions in the "
                f"token layout ({na_dim}, got {q_tile_shape=}, {kv_tile_shape=}."
            )

        if any(x % t != 0 for x, t in zip(q_shape, q_tile_shape)):
            raise ValueError(
                "Input must be divisible by Q tile shape, but got "
                f"{q_shape=}, {q_tile_shape=}."
            )

        if any(x % t != 0 for x, t in zip(kv_shape, kv_tile_shape)):
            raise ValueError(
                "Input must be divisible by KV tile shape, but got "
                f"{kv_shape=}, {kv_tile_shape=}."
            )

        q_rest_shape = tuple(x // t for x, t in zip(q_shape, q_tile_shape))
        kv_rest_shape = tuple(x // t for x, t in zip(kv_shape, kv_tile_shape))

    def single_dim_tiling_mask(
        b: IntTensor,
        h: IntTensor,
        q_idx: IntTensor,
        kv_idx: IntTensor,
        qkv_shape,
        kernel_size,
        stride,
        dilation,
        is_causal,
    ) -> BoolTensor:

        # Reconstruct global Q and KV coordinates
        q_crd = idx2crd(q_idx, qkv_shape)
        kv_crd = idx2crd(kv_idx, qkv_shape)

        # Coordinates within dilation group
        q_crd_di = tuple(x // d for x, d in zip(q_crd, dilation))
        kv_crd_di = tuple(x // d for x, d in zip(kv_crd, dilation))

        # Dilation group coordinates
        q_dilation_group_crd = tuple(x % d for x, d in zip(q_crd, dilation))
        kv_dilation_group_crd = tuple(x % d for x, d in zip(kv_crd, dilation))

        # Fixup input shape according to dilation group
        dilation_group_padding = tuple(
            1 - ((dg + (d - (x % d))) // d)
            for dg, d, x in zip(q_dilation_group_crd, dilation, qkv_shape)
        )
        qkv_shape_corrected = tuple(
            (x // d) + p for p, d, x in zip(dilation_group_padding, dilation, qkv_shape)
        )

        # Window size left and right (non-causal only)
        window_size_left = tuple(w // 2 for w in kernel_size)
        window_size_right = tuple(w // 2 + (w % 2 - 1) for w in kernel_size)

        masks = []
        for i in range(na_dim):
            if is_causal[i]:
                # Leader is the last (right-most) query in the stride group.
                stride_group_leader = torch.min(
                    (q_crd_di[i] // stride[i]) * stride[i] + stride[i] - 1,
                    qkv_shape_corrected[i] - 1,
                )

                mask = (
                    (
                        q_crd_di[i] - kv_crd_di[i] >= 0
                    )  # window still ends at query index
                    & (stride_group_leader - kv_crd_di[i] < kernel_size[i])
                    & (q_dilation_group_crd[i] == kv_dilation_group_crd[i])
                )
            else:
                # Leader is the center-most query in the stride group.
                # If stride is even, choose the right hand side center query.
                stride_group_leader = torch.min(
                    (q_crd_di[i] // stride[i]) * stride[i] + (stride[i] // 2),
                    qkv_shape_corrected[i] - 1,
                )

                window_center = stride_group_leader.clamp(
                    window_size_left[i] * torch.ones_like(qkv_shape_corrected[i]),
                    qkv_shape_corrected[i] - 1 - window_size_right[i],
                )
                w0 = window_center - kv_crd_di[i]
                w1 = kv_crd_di[i] - window_center
                mask = (
                    ((0 <= w0) & (w0 <= window_size_left[i]))
                    | ((0 <= w1) & (w1 <= window_size_right[i]))
                ) & (q_dilation_group_crd[i] == kv_dilation_group_crd[i])

            masks.append(mask)

        return functools.reduce(lambda x, y: x & y, masks)  # type: ignore

    def multi_dim_tiling_mask(
        b: IntTensor,
        h: IntTensor,
        q_idx: IntTensor,
        kv_idx: IntTensor,
        q_tile_size: int,
        kv_tile_size: int,
        q_tile_shape,
        kv_tile_shape,
        qkv_shape,
        kernel_size,
        stride,
        dilation,
        is_causal,
    ) -> BoolTensor:

        # Reconstruct global Q and KV coordinates
        q_tile_idx = q_idx // q_tile_size
        kv_tile_idx = kv_idx // kv_tile_size
        q_tile_offset = q_idx % q_tile_size
        kv_tile_offset = kv_idx % q_tile_size
        q_tile_coord = idx2crd(q_tile_idx, q_rest_shape)
        kv_tile_coord = idx2crd(kv_tile_idx, kv_rest_shape)
        q_tile_offset_coord = idx2crd(q_tile_offset, q_tile_shape)
        kv_tile_offset_coord = idx2crd(kv_tile_offset, kv_tile_shape)

        q_crd = tuple(
            tile_crd * tile_sz + tile_off
            for tile_crd, tile_sz, tile_off in zip(
                q_tile_coord, q_tile_shape, q_tile_offset_coord
            )
        )
        kv_crd = tuple(
            tile_crd * tile_sz + tile_off
            for tile_crd, tile_sz, tile_off in zip(
                kv_tile_coord, kv_tile_shape, kv_tile_offset_coord
            )
        )

        # Dilation group coordinates
        # h_actual = h % num_heads
        dilation_group_idx = h // num_heads
        dilation_group_crd = idx2crd(dilation_group_idx, dilation)

        # Fixup input shape according to dilation group
        dilation_group_padding = tuple(
            1 - ((dg + (d - (x % d))) // d)
            for dg, d, x in zip(dilation_group_crd, dilation, qkv_shape)
        )
        qkv_shape_corrected = tuple(
            (x // d) + p for p, d, x in zip(dilation_group_padding, dilation, qkv_shape)
        )

        # Window size left and right (non-causal only)
        window_size_left = tuple(w // 2 for w in kernel_size)
        window_size_right = tuple(w // 2 + (w % 2 - 1) for w in kernel_size)

        masks = []
        for i in range(na_dim):
            if is_causal[i]:
                # Leader is the last (right-most) query in the stride group.
                stride_group_leader = torch.min(
                    (q_crd[i] // stride[i]) * stride[i] + stride[i] - 1,
                    qkv_shape_corrected[i] - 1,
                )

                mask = (
                    q_crd[i] - kv_crd[i] >= 0
                ) & (  # window still ends at query index
                    stride_group_leader - kv_crd[i] < kernel_size[i]
                )
            else:
                # Leader is the center-most query in the stride group.
                # If stride is even, choose the right hand side center query.
                stride_group_leader = torch.min(
                    (q_crd[i] // stride[i]) * stride[i] + (stride[i] // 2),
                    qkv_shape_corrected[i] - 1,
                )

                window_center = stride_group_leader.clamp(
                    window_size_left[i] * torch.ones_like(qkv_shape_corrected[i]),
                    qkv_shape_corrected[i] - 1 - window_size_right[i],
                )
                w0 = window_center - kv_crd[i]
                w1 = kv_crd[i] - window_center

                mask = ((0 <= w0) & (w0 <= window_size_left[i])) | (
                    (0 <= w1) & (w1 <= window_size_right[i])
                )

            masks.append(mask)

        return functools.reduce(lambda x, y: x & y, masks)  # type: ignore

    mask_mod = None
    seq_length_q = seq_length_kv = math.prod(qkv_shape)
    q_tile_size, kv_tile_size = 64, 64
    if do_token_permute:
        assert q_shape is not None
        assert kv_shape is not None
        assert q_tile_shape is not None
        assert kv_tile_shape is not None

        seq_length_q = math.prod(q_shape)
        seq_length_kv = math.prod(kv_shape)
        q_tile_size, kv_tile_size = math.prod(q_tile_shape), math.prod(kv_tile_shape)

        mask_mod = functools.partial(
            multi_dim_tiling_mask,
            q_tile_size=q_tile_size,
            kv_tile_size=kv_tile_size,
            q_tile_shape=q_tile_shape,
            kv_tile_shape=kv_tile_shape,
            qkv_shape=qkv_shape,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            is_causal=is_causal,
        )
    else:
        mask_mod = functools.partial(
            single_dim_tiling_mask,
            qkv_shape=qkv_shape,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            is_causal=is_causal,
        )

    block_mask = create_block_mask(
        mask_mod,
        B=None,
        H=None,
        Q_LEN=seq_length_q,
        KV_LEN=seq_length_kv,
        _compile=torch_compile,
        BLOCK_SIZE=(q_tile_size, kv_tile_size),
        device=device,
    )
    flex_mask_end_time = time.perf_counter()
    flex_mask_time = flex_mask_end_time - flex_mask_start_time
    logger.debug(
        f"Flex Attention block mask ({torch_compile=}) created in {flex_mask_time:.2f} seconds."
    )
    return block_mask


def flex_fna_generic(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: DimensionTypeOrDed,
    stride: DimensionTypeOrDed = 1,
    dilation: DimensionTypeOrDed = 1,
    is_causal: Optional[CausalArgTypeOrDed] = False,
    scale: Optional[float] = None,
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
    torch_compile: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:

    na_tensor_checks(query, key, value, must_match_head_dims=True)

    na_dim = query.dim() - 3  # batch, heads, head_dim

    kernel_size, stride, dilation, is_causal = check_all_args(
        na_dim, kernel_size, stride, dilation, is_causal
    )

    check_args_against_input(
        query,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
    )

    batch_size, *qkv_shape_in, num_heads, head_dim = query.shape
    qkv_shape = check_input_size_arg(na_dim, qkv_shape_in)

    scale = scale or query.shape[-1] ** -0.5

    assert can_run_flex_attention(
        query, key, value, torch_compile=torch_compile, raise_error=True
    )

    if (q_tile_shape is None) ^ (kv_tile_shape is None):
        raise ValueError(
            "Please specify both q_tile_shape and kv_tile_shape, or neither one. "
            f"Got {q_tile_shape=}, {kv_tile_shape=}."
        )

    do_token_permute = q_tile_shape is not None and kv_tile_shape is not None

    q_shape = kv_shape = qkv_shape
    q_tile_size: Optional[int] = None
    kv_tile_size: Optional[int] = None
    if do_token_permute:
        q_tile_shape, kv_tile_shape = check_flex_fna_forward_config(
            input_tensor=query,
            q_tile_shape=q_tile_shape,
            kv_tile_shape=kv_tile_shape,
        )

        q_tile_size = math.prod(q_tile_shape)
        kv_tile_size = math.prod(kv_tile_shape)

        query_pad, padding = maybe_pad(query, q_tile_shape, dilation=dilation)
        key, _ = maybe_pad(key, kv_tile_shape, dilation=dilation)
        value, _ = maybe_pad(value, kv_tile_shape, dilation=dilation)

        query_perm, q_shape, qR = token_permute(
            query_pad, q_tile_shape, dilation=dilation, flip_tiled_dims=False
        )
        key_perm, k_shape, kR = token_permute(
            key, kv_tile_shape, dilation=dilation, flip_tiled_dims=False
        )
        value_perm, v_shape, vR = token_permute(
            value, kv_tile_shape, dilation=dilation, flip_tiled_dims=False
        )

        assert k_shape == v_shape
        kv_shape = k_shape

        # Token permute already flattens to 1-D
        # Flex uses heads first layout
        query_ = query_perm.transpose(1, 2)
        key_ = key_perm.transpose(1, 2)
        value_ = value_perm.transpose(1, 2)

    else:
        seqlen = math.prod(qkv_shape)
        # Flex uses heads first layout
        query_ = query.reshape(batch_size, seqlen, num_heads, head_dim).transpose(1, 2)
        key_ = key.reshape(batch_size, seqlen, num_heads, head_dim).transpose(1, 2)
        value_ = value.reshape(batch_size, seqlen, num_heads, head_dim).transpose(1, 2)

    na_block_mask = get_na_flex_mask(
        device=query.device.type,
        na_dim=na_dim,
        num_heads=num_heads,
        qkv_shape=qkv_shape,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        q_shape=q_shape,
        kv_shape=kv_shape,
        torch_compile=torch_compile,
    )

    out_, lse_ = run_flex_attn(
        query_,
        key_,
        value_,
        na_block_mask,
        scale,
        torch_compile=torch_compile,
        q_tile_size=q_tile_size,
        kv_tile_size=kv_tile_size,
    )

    if do_token_permute:
        out = out_.transpose(1, 2)
        lse = lse_.transpose(1, 2).unsqueeze(-1)

        out = maybe_unpad(
            token_unpermute(
                out, q_tile_shape, q_shape, qR, dilation=dilation, flip_tiled_dims=False
            ),
            padding,
        )
        lse = maybe_unpad(
            token_unpermute(
                lse, q_tile_shape, q_shape, qR, dilation=dilation, flip_tiled_dims=False
            ),
            padding,
        ).squeeze(-1)
    else:
        out = out_.transpose(1, 2).reshape(batch_size, *qkv_shape, num_heads, head_dim)
        lse = lse_.transpose(1, 2).reshape(batch_size, *qkv_shape, num_heads)

    if return_lse:
        return out, lse

    return out


def na1d_flex(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension1DTypeOrDed,
    stride: Dimension1DTypeOrDed = 1,
    dilation: Dimension1DTypeOrDed = 1,
    is_causal: Optional[CausalArg1DTypeOrDed] = False,
    scale: Optional[float] = None,
    q_tile_shape: Optional[Dimension1DType] = None,
    kv_tile_shape: Optional[Dimension1DType] = None,
    torch_compile: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    return flex_fna_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        torch_compile=torch_compile,
        return_lse=return_lse,
    )


def na2d_flex(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension2DTypeOrDed,
    stride: Dimension2DTypeOrDed = 1,
    dilation: Dimension2DTypeOrDed = 1,
    is_causal: Optional[CausalArg2DTypeOrDed] = False,
    scale: Optional[float] = None,
    q_tile_shape: Optional[Dimension2DType] = None,
    kv_tile_shape: Optional[Dimension2DType] = None,
    torch_compile: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    return flex_fna_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        torch_compile=torch_compile,
        return_lse=return_lse,
    )


def na3d_flex(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension3DTypeOrDed,
    stride: Dimension3DTypeOrDed = 1,
    dilation: Dimension3DTypeOrDed = 1,
    is_causal: Optional[CausalArg3DTypeOrDed] = False,
    scale: Optional[float] = None,
    q_tile_shape: Optional[Dimension3DType] = None,
    kv_tile_shape: Optional[Dimension3DType] = None,
    torch_compile: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    return flex_fna_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        torch_compile=torch_compile,
        return_lse=return_lse,
    )



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################


from typing import Dict, List

from ....types import QKTileShapeType

# TODO: More combinations are possible for
# 2D and 3D (query tile does not have to be smaller
# than KV tile); but that behavior is untested,
# and IIRC was unstable.

_FNA_FORWARD_64x128_TILE_SIZES: Dict[int, List[QKTileShapeType]] = {
    1: [
        ((64,), (128,)),
    ],
    2: [
        ((32, 2), (64, 2)),
        ((32, 2), (32, 4)),
        ((16, 4), (32, 4)),
        ((16, 4), (16, 8)),
        ((8, 8), (16, 8)),
        ((8, 8), (8, 16)),
        ((4, 16), (8, 16)),
        ((4, 16), (4, 32)),
        ((2, 32), (4, 32)),
        ((2, 32), (2, 64)),
    ],
    3: [
        ((16, 2, 2), (32, 2, 2)),
        ((16, 2, 2), (16, 4, 2)),
        ((16, 2, 2), (16, 2, 4)),
        ((8, 4, 2), (16, 4, 2)),
        ((8, 4, 2), (8, 8, 2)),
        ((8, 4, 2), (8, 4, 4)),
        ((8, 2, 4), (16, 2, 4)),
        ((8, 2, 4), (8, 4, 4)),
        ((8, 2, 4), (8, 2, 8)),
        ((4, 8, 2), (8, 8, 2)),
        ((4, 8, 2), (4, 16, 2)),
        ((4, 8, 2), (4, 8, 4)),
        ((4, 4, 4), (8, 4, 4)),
        ((4, 4, 4), (4, 8, 4)),
        ((4, 4, 4), (4, 4, 8)),
        ((4, 2, 8), (8, 2, 8)),
        ((4, 2, 8), (4, 4, 8)),
        ((4, 2, 8), (4, 2, 16)),
        ((2, 16, 2), (4, 16, 2)),
        ((2, 16, 2), (2, 32, 2)),
        ((2, 16, 2), (2, 16, 4)),
        ((2, 8, 4), (4, 8, 4)),
        ((2, 8, 4), (2, 16, 4)),
        ((2, 8, 4), (2, 8, 8)),
        ((2, 4, 8), (4, 4, 8)),
        ((2, 4, 8), (2, 8, 8)),
        ((2, 4, 8), (2, 4, 16)),
        ((2, 2, 16), (4, 2, 16)),
        ((2, 2, 16), (2, 4, 16)),
        ((2, 2, 16), (2, 2, 32)),
    ],
}



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from typing import List, Optional

import torch  # noqa: F401
from torch import Tensor

from ....types import DimensionType, FlexFmhaForwardConfigType, FlexFnaForwardConfigType
from ....utils.checks import check_tile_shape
from ....utils.device import get_device_cc


# TODO: add more tile sizes/shapes
# TODO: add backprop tile sizes/shapes
# Only doing 64 x 64 for now, since it's the one that successfully compiles across devices and
# use cases without running into compile errors (i.e. shmem over-subscription)
# Once Flex with compilation actually starts working as expected and is out of prototype, we can
# add in more tile sizes/shapes and condition them on arch / use case, like we do for CUTLASS FNA.

FLEX_FORWARD_TILE_SHAPES = {
    1: [
        # ((128, ), (128, )),
        ((64,), (64,)),
    ],
    2: [
        # ((8, 16), (8, 16)),
        ((8, 8), (8, 8)),
        ((4, 16), (4, 16)),
        ((4, 16), (8, 8)),
    ],
    3: [
        # ((4, 4, 8), (4, 4, 8)),
        ((4, 4, 4), (4, 4, 4)),
        ((2, 4, 8), (2, 4, 8)),
        ((2, 4, 8), (4, 4, 4)),
    ],
}


def _get_default_tile_shapes_forward(
    na_dim: int,
) -> FlexFnaForwardConfigType:
    assert na_dim in [1, 2, 3]

    if na_dim == 1:
        return ((64,), (64,))
    if na_dim == 2:
        return ((8, 8), (8, 8))
    if na_dim == 3:
        return ((4, 4, 4), (4, 4, 4))

    raise NotImplementedError()


def get_all_tile_shapes_forward(
    input_tensor: Tensor,
) -> List[FlexFnaForwardConfigType]:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim

    return FLEX_FORWARD_TILE_SHAPES[na_dim]  # type: ignore


# For FMHA
def get_all_tile_sizes_forward(input_tensor: Tensor) -> List[FlexFmhaForwardConfigType]:
    if input_tensor.dim() != 4:
        raise ValueError("Only 4-D tensors are supported in FMHA.")

    tile_shapes = get_all_tile_shapes_forward(input_tensor)
    assert all(len(q_t) == len(kv_t) == 1 for q_t, kv_t in tile_shapes)

    tile_sizes = [(q_t[0], kv_t[0]) for q_t, kv_t in tile_shapes]

    return tile_sizes


def get_default_forward_tile_shapes(input_tensor: Tensor) -> FlexFnaForwardConfigType:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim

    return _get_default_tile_shapes_forward(na_dim)


def get_default_forward_tile_sizes(input_tensor: Tensor) -> FlexFmhaForwardConfigType:
    if input_tensor.dim() != 4:
        raise ValueError("Only 4-D tensors are supported in FMHA.")

    q_t, kv_t = get_default_forward_tile_shapes(input_tensor)
    assert len(q_t) == len(kv_t) == 1

    return (q_t[0], kv_t[0])


def check_flex_fna_forward_config(
    input_tensor: Tensor,
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
) -> FlexFnaForwardConfigType:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim

    if (q_tile_shape is None) ^ (kv_tile_shape is None):
        raise ValueError(
            "Please specify both q_tile_shape and kv_tile_shape, or neither one. "
            f"Got {q_tile_shape=}, {kv_tile_shape=}."
        )

    if q_tile_shape is None and kv_tile_shape is None:
        return get_default_forward_tile_shapes(input_tensor=input_tensor)

    q_tile_shape = check_tile_shape(q_tile_shape)
    kv_tile_shape = check_tile_shape(kv_tile_shape)

    tile_shapes = get_all_tile_shapes_forward(input_tensor=input_tensor)

    for q_t, kv_t in tile_shapes:
        if q_t == q_tile_shape and kv_t == kv_tile_shape:
            return (q_t, kv_t)  # type: ignore

    # Fail and make suggestions
    device_cc = get_device_cc(input_tensor.device)
    MAX_EXAMPLES = 3
    examples = ""
    for i, (q_t, kv_t) in enumerate(tile_shapes):
        examples += f"\n  q_tile_shape={q_t}, kv_tile_shape={kv_t}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for Flex FNA-{na_dim}D. "
        f"Q tile shape {q_tile_shape} and KV tile shape {kv_tile_shape} "
        f"are not among the {len(tile_shapes)} configurations implementable "
        f"with Flex FNA for SM{device_cc}. "
        "Try selecting a combination from: \n"
        "  natten.get_configs_for_flex_fna(q, k, v)"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )


def check_flex_fmha_forward_config(
    input_tensor: Tensor,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
) -> FlexFmhaForwardConfigType:
    assert input_tensor.dim() == 4

    if (q_tile_size is None) ^ (kv_tile_size is None):
        raise ValueError(
            "Please specify both q_tile_size and kv_tile_size, or neither one. "
            f"Got {q_tile_size=}, {kv_tile_size=}."
        )

    if q_tile_size is None and kv_tile_size is None:
        q_tile_size, kv_tile_size = get_default_forward_tile_sizes(
            input_tensor=input_tensor
        )
        return (q_tile_size, kv_tile_size)

    tile_sizes = get_all_tile_sizes_forward(input_tensor=input_tensor)

    for q_t, kv_t in tile_sizes:
        if q_t == q_tile_size and kv_t == kv_tile_size:
            return (q_t, kv_t)

    # Fail and make suggestions
    device_cc = get_device_cc(input_tensor.device)
    MAX_EXAMPLES = 3
    examples = ""
    for i, (q_t, kv_t) in enumerate(tile_sizes):
        examples += f"\n  q_tile_size={q_t}, kv_tile_size={kv_t}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for Flex FMHA. "
        f"Q tile size {q_tile_size} and KV tile size {kv_tile_size} "
        f"are not among the {len(tile_sizes)} configurations implementable "
        f"with Flex FMHA for SM{device_cc}. "
        "Try selecting a combination from: \n"
        "  natten.get_configs_for_flex_fmha(q, k, v)\n"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import torch  # noqa: F401

try:
    from natten import libnatten  # noqa: F401

    from natten.libnatten import (  # type: ignore[import-untyped]
        blackwell_fmha_backward,
        blackwell_fmha_forward,
        blackwell_na1d_backward,
        blackwell_na1d_forward,
        blackwell_na2d_backward,
        blackwell_na2d_forward,
        blackwell_na3d_backward,
        blackwell_na3d_forward,
        compute_delta,
        fmha_backward,
        fmha_forward,
        hopper_fmha_backward,
        hopper_fmha_forward,
        hopper_na1d_backward,
        hopper_na1d_forward,
        hopper_na2d_backward,
        hopper_na2d_forward,
        hopper_na3d_backward,
        hopper_na3d_forward,
        na1d_backward,
        na1d_forward,
        na2d_backward,
        na2d_forward,
        na3d_backward,
        na3d_forward,
        reference_na1d_backward,
        reference_na1d_forward,
        reference_na2d_backward,
        reference_na2d_forward,
        reference_na3d_backward,
        reference_na3d_forward,
    )

    HAS_LIBNATTEN = True

except ImportError:
    HAS_LIBNATTEN = False

    # Stubs
    def libnatten_import_error():
        raise ImportError(
            "Failed to import libnatten. "
            "Did you build / download NATTEN with the CUDA backend (libnatten)? "
            "Please make sure you built NATTEN correctly, or refer to "
            "https://shi-labs.com/natten for more information."
        )

    # Reference kernels
    def reference_na1d_forward(*args, **kwargs):
        libnatten_import_error()

    def reference_na2d_forward(*args, **kwargs):
        libnatten_import_error()

    def reference_na3d_forward(*args, **kwargs):
        libnatten_import_error()

    def reference_na1d_backward(*args, **kwargs):
        libnatten_import_error()

    def reference_na2d_backward(*args, **kwargs):
        libnatten_import_error()

    def reference_na3d_backward(*args, **kwargs):
        libnatten_import_error()

    # CUTLASS 2.X kernels
    def fmha_forward(*args, **kwargs):
        libnatten_import_error()

    def fmha_backward(*args, **kwargs):
        libnatten_import_error()

    ## SM50/SM70/SM75/SM80 - Original FNA
    def na1d_forward(*args, **kwargs):
        libnatten_import_error()

    def na2d_forward(*args, **kwargs):
        libnatten_import_error()

    def na3d_forward(*args, **kwargs):
        libnatten_import_error()

    def na1d_backward(*args, **kwargs):
        libnatten_import_error()

    def na2d_backward(*args, **kwargs):
        libnatten_import_error()

    def na3d_backward(*args, **kwargs):
        libnatten_import_error()

    # CUTLASS 3.X kernels
    ## SM90 - Hopper FMHA
    def hopper_fmha_forward(*args, **kwargs):
        libnatten_import_error()

    def hopper_fmha_backward(*args, **kwargs):
        libnatten_import_error()

    ## SM90 - Hopper FNA
    def hopper_na1d_forward(*args, **kwargs):
        libnatten_import_error()

    def hopper_na2d_forward(*args, **kwargs):
        libnatten_import_error()

    def hopper_na3d_forward(*args, **kwargs):
        libnatten_import_error()

    def hopper_na1d_backward(*args, **kwargs):
        libnatten_import_error()

    def hopper_na2d_backward(*args, **kwargs):
        libnatten_import_error()

    def hopper_na3d_backward(*args, **kwargs):
        libnatten_import_error()

    ## SM100 - Blackwell FMHA
    def blackwell_fmha_forward(*args, **kwargs):
        libnatten_import_error()

    def blackwell_fmha_backward(*args, **kwargs):
        libnatten_import_error()

    ## SM100 - Blackwell FNA
    def blackwell_na1d_forward(*args, **kwargs):
        libnatten_import_error()

    def blackwell_na2d_forward(*args, **kwargs):
        libnatten_import_error()

    def blackwell_na3d_forward(*args, **kwargs):
        libnatten_import_error()

    def blackwell_na1d_backward(*args, **kwargs):
        libnatten_import_error()

    def blackwell_na2d_backward(*args, **kwargs):
        libnatten_import_error()

    def blackwell_na3d_backward(*args, **kwargs):
        libnatten_import_error()

    # Misc kernels
    def compute_delta(*args, **kwargs):
        libnatten_import_error()
